# -*- coding: utf-8 -*-
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""ProtocolError and StandardClient implementation"""

import socket
import sys

import gevent
import gevent.lock

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class TimeoutError(Exception):
    """Protype"""


class ProtocolError(Exception):
    """Protype"""


class SocketError(Exception):
    """Protype"""


if sys.version_info > (3, 0):
    STX = 2  # b'\x02'
    ETX = 3  # b'\x03'

    def empty_buffer():
        """Empty buffer"""
        return b""

    _bytes = bytes

    encode = str.encode

else:
    STX = chr(2)
    ETX = chr(3)

    def empty_buffer():
        """Empty buffer"""
        return ""

    _bytes = str

    encode = str

MAX_SIZE_STREAM_MSG = 500000


class PROTOCOL:
    """Protocol"""

    DATAGRAM = 1
    STREAM = 2


class StandardClient:
    """Standard JLib client"""

    def __init__(self, server_ip, server_port, protocol, timeout, retries):
        self.server_ip = server_ip
        self.server_port = server_port
        self.timeout = timeout
        self.default_timeout = timeout
        self.retries = retries
        self.protocol = protocol
        self.error = None
        self.received_msg = None
        self.receiving_greenlet = None
        self.msg_received_event = gevent.event.Event()
        self._lock = gevent.lock.Semaphore()
        self.__msg_index__ = -1
        self.__sock = None
        self.__constant_local_port = True
        self._is_connected = False

    def __create_socket(self):
        """Create socket"""
        if self.protocol == PROTOCOL.DATAGRAM:
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.__sock.settimeout(self.timeout)
        else:
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def __close_socket(self):
        """Close socket"""
        try:
            self.__sock.close()
        except Exception:
            pass
        self._is_connected = False
        self.__sock = None
        self.received_msg = None

    def connect(self):
        """Socket connect"""
        if self.protocol == PROTOCOL.DATAGRAM:
            return
        if self.__sock is None:
            self.__create_socket()

        self.__sock.connect((self.server_ip, self.server_port))
        self._is_connected = True
        self.error = None
        self.received_msg = None
        self.receiving_greenlet = gevent.spawn(self.recv_thread)

    def is_connected(self):
        """Check if connected
        Returns:
            (bool): True if connected
        """
        if self.protocol == PROTOCOL.DATAGRAM:
            return False
        if self.__sock is None:
            return False
        return self._is_connected

    def disconnect(self):
        """Disconnect"""
        if self.is_connected():
            self.receiving_greenlet.kill()
        self.__close_socket()

    def __send_receive_datagram_single(self, cmd):
        """Send and receive single datagram.
        Args:
            cmd(str): Command
        Returns:
            (str): return message
        Raises:
            SocketError, TimeoutError
        """
        try:
            if self.__constant_local_port is False or self.__sock is None:
                self.__create_socket()
            msg_number = "%04d " % self.__msg_index__
            msg = msg_number + cmd
            try:
                self.__sock.sendto(encode(msg), (self.server_ip, self.server_port))
            except Exception:
                raise SocketError("Socket error:" + str(sys.exc_info()[1]))
            received = False
            while received is False:
                try:
                    ret = self.__sock.recv(4096).decode()
                except socket.timeout:
                    raise TimeoutError("Timeout error:" + str(sys.exc_info()[1]))
                except Exception:
                    raise SocketError("Socket error:" + str(sys.exc_info()[1]))
                if ret[0:5] == msg_number:
                    received = True
            ret = ret[5:]
        except SocketError:
            self.__close_socket()
            raise
        except Exception:
            if self.__constant_local_port is False:
                self.__close_socket()
            raise
        if self.__constant_local_port is False:
            self.__close_socket()
        return ret

    def __send_receive_datagram(self, cmd):
        """Send/receive datagram.
        Args:
            (str): command
        Returns:
            (str): datagram
        Raises:
            TimeoutError, ProtocolError
        """
        self.__msg_index__ = self.__msg_index__ + 1
        if self.__msg_index__ >= 10000:
            self.__msg_index__ = 1
        for i in range(0, self.retries):
            try:
                ret = self.__send_receive_datagram_single(encode(cmd))
                return ret
            except TimeoutError:
                if i >= self.retries - 1:
                    raise
            except ProtocolError:
                if i >= self.retries - 1:
                    raise
            except SocketError:
                if i >= self.retries - 1:
                    raise
            except Exception:
                raise

    def set_timeout(self, timeout):
        """Set the socket timeout.
        Args:
            timeout(float): Timeout value
        """
        self.timeout = timeout
        if self.protocol == PROTOCOL.DATAGRAM:
            if self.__sock is not None:
                self.__sock.settimeout(self.timeout)

    def restore_timeout(self):
        """Restore the default timeout"""
        self.set_timeout(self.default_timeout)

    def dispose(self):
        """Disconnect or close socket"""
        if self.protocol == PROTOCOL.DATAGRAM:
            if self.__constant_local_port:
                self.__close_socket()
            else:
                pass
        else:
            self.disconnect()

    def on_message_received(self, msg):
        """Actions
        Args:
            msg(str): Message
        """
        self.received_msg = msg
        self.msg_received_event.set()

    def recv_thread(self):
        """Receive thread"""
        try:
            self.on_connected()
        except Exception:
            pass
        buffer = empty_buffer()
        mReceivedSTX = False
        while True:
            ret = self.__sock.recv(4096)
            if not ret:
                # connection reset by peer
                self.error = "Disconnected"
                self.__close_socket()
                break
            for b in ret:
                if b == STX:
                    buffer = empty_buffer()
                    mReceivedSTX = True
                elif b == ETX:
                    if mReceivedSTX:
                        try:
                            # Unicode decoding exception catching,
                            # consider errors='ignore'
                            buffer_utf8 = buffer.decode()
                        except UnicodeDecodeError as e:
                            # Syntax not allowed in Python 2
                            # raise ProtocolError from e
                            raise ProtocolError(
                                "UnicodeDecodeError: %s" % (sys.exc_info(),)
                            )
                        self.on_message_received(buffer_utf8)
                        mReceivedSTX = False
                        buffer = empty_buffer()
                else:
                    if mReceivedSTX:
                        buffer += _bytes([b])

            if len(buffer) > MAX_SIZE_STREAM_MSG:
                mReceivedSTX = False
                buffer = empty_buffer()
        try:
            self.on_disconnected()
        except Exception:
            pass

    def __send_stream(self, cmd):
        """Send a command.
        Args:
            cmd(str): command
        """
        if not self.is_connected():
            self.connect()
        try:
            pack = _bytes([STX]) + encode(cmd) + _bytes([ETX])
            self.__sock.send(pack)
        except SocketError:
            self.disconnect()

    def __send_receive_stream(self, cmd):
        """Send/receive event.
        Args:
            cmd(str): command
        Returns:
            (str): reply form the socket
        """
        self.error = None
        self.received_msg = None
        self.msg_received_event.clear()  # = gevent.event.Event()
        if not self.is_connected():
            self.connect()
        self.__send_stream(cmd)

        with gevent.Timeout(self.timeout, TimeoutError):
            while self.received_msg is None:
                if self.error is not None:
                    raise SocketError("Socket error:" + str(self.error))
                self.msg_received_event.wait()
            return self.received_msg

    def send_receive(self, cmd, timeout=-1):
        """Send/receive command, locking the socket.
        Args:
            cmd(str): command
        Returns:
            (str): reply form the socket
        """
        self._lock.acquire()
        try:
            if (timeout is None) or (timeout >= 0):
                self.set_timeout(timeout)
            if self.protocol == PROTOCOL.DATAGRAM:
                return self.__send_receive_datagram(cmd)
            return self.__send_receive_stream(cmd)
        finally:
            try:
                if (timeout is None) or (timeout >= 0):
                    self.restore_timeout()
            finally:
                self._lock.release()

    def send(self, cmd):
        """Send command.
        Args:
            cmd(str): command
        Returns:
            (str): reply form the socket
        Raises:
            ProtocolError
        """
        if self.protocol == PROTOCOL.DATAGRAM:
            raise ProtocolError(
                "Protocol error: send command not support in datagram clients"
            )
        return self.__send_stream(cmd)

    def on_connected(self):
        """On connect"""

    def on_disconnected(self):
        """On disconnect"""



"""
[Name] BeamInfo

[Description]
BeamInfo hardware object is used to define final beam size and shape.
It can include aperture, slits and/or other beam definer (lenses or other eq.)

[Emitted signals]
beamInfoChanged
beamPosChanged

[Included Hardware Objects]
-----------------------------------------------------------------------
| name            | signals          | functions
-----------------------------------------------------------------------
 aperture_hwobj	    apertureChanged
 slits_hwobj
 beam_definer_hwobj
-----------------------------------------------------------------------
"""

import logging

from mxcubecore.BaseHardwareObjects import HardwareObject


class ALBABeamInfo(HardwareObject):
    """
    Description:
    """

    def __init__(self, *args):
        """
        Descrip. :
        """
        super().__init__(*args)

        self.aperture_hwobj = None
        self.slits_hwobj = None

        self.beam_size_slits = None
        self.beam_size_aperture = None
        self.beam_size_definer = None
        self.beam_position = None
        self.beam_info_dict = None
        self.default_beam_divergence = None

    def init(self):
        """
        Descript. :
        """
        self.beam_size_slits = [9999, 9999]
        self.beam_size_aperture = [9999, 9999]
        self.beam_size_definer = [9999, 9999]

        self.beam_position = [0, 0]
        self.beam_info_dict = {}

        self.beam_width_chan = self.get_channel_object("BeamWidth")
        self.beam_height_chan = self.get_channel_object("BeamHeight")
        self.beam_posx_chan = self.get_channel_object("BeamPositionHorizontal")
        self.beam_posy_chan = self.get_channel_object("BeamPositionVertical")

        self.beam_height_chan.connect_signal("update", self.beam_height_changed)
        self.beam_width_chan.connect_signal("update", self.beam_width_changed)
        self.beam_posx_chan.connect_signal("update", self.beam_posx_changed)
        self.beam_posy_chan.connect_signal("update", self.beam_posy_changed)

        # divergence can be added as fixed properties in xml
        default_beam_divergence_vertical = None
        default_beam_divergence_horizontal = None

        try:
            default_beam_divergence_vertical = int(
                self.get_property("beam_divergence_vertical")
            )
            default_beam_divergence_horizontal = int(
                self.get_property("beam_divergence_horizontal")
            )
        except Exception:
            pass

        self.default_beam_divergence = [
            default_beam_divergence_horizontal,
            default_beam_divergence_vertical,
        ]

    def connect_notify(self, *args):
        self.evaluate_beam_info()
        self.re_emit_values()

    def get_beam_divergence_hor(self):
        """
        Descript. :
        """
        return self.default_beam_divergence[0]

    def get_beam_divergence_ver(self):
        """
        Descript. :
        """
        return self.default_beam_divergence[1]

    def get_beam_position(self):
        """
        Descript. :
        Arguments :
        Return    :
        """
        self.beam_position = (
            self.beam_posx_chan.get_value(),
            self.beam_posy_chan.get_value(),
        )
        return self.beam_position

    def get_slits_gap(self):
        return None, None

    def set_beam_position(self, beam_x, beam_y):
        """
        Descript. :
        Arguments :
        Return    :
        """
        self.beam_position = (beam_x, beam_y)

    def get_beam_info(self):
        """
        Descript. :
        Arguments :
        Return    :
        """
        return self.evaluate_beam_info()

    def get_beam_size(self):
        """
        Descript. : returns beam size in millimeters
        Return   : list with two integers
        """
        self.evaluate_beam_info()
        return self.beam_info_dict["size_x"], self.beam_info_dict["size_y"]

    def get_beam_shape(self):
        """
        Descript. :
        Arguments :
        Return    :
        """
        return self.beam_info_dict["shape"]

    def beam_width_changed(self, value):
        self.beam_info_dict["size_x"] = value
        self.re_emit_values()

    def beam_height_changed(self, value):
        self.beam_info_dict["size_y"] = value
        self.re_emit_values()

    def beam_posx_changed(self, value):
        self.beam_position["x"] = value
        self.re_emit_values()

    def beam_posy_changed(self, value):
        self.beam_position["y"] = value
        self.re_emit_values()

    def evaluate_beam_info(self):
        """
        Descript. : called if aperture, slits or focusing has been changed
        Return    : dictionary,{size_x:0.1, size_y:0.1, shape:"rectangular"}
        """

        self.beam_info_dict["size_x"] = self.beam_width_chan.get_value() / 1000.0
        self.beam_info_dict["size_y"] = self.beam_height_chan.get_value() / 1000.0
        self.beam_info_dict["shape"] = "rectangular"

        return self.beam_info_dict

    def re_emit_values(self):
        """
        Descript. :
        Arguments :
        Return    :
        """
        logging.getLogger("HWR").debug(" emitting beam info")
        if (
            self.beam_info_dict["size_x"] != 9999
            and self.beam_info_dict["size_y"] != 9999
        ):
            self.emit(
                "beamSizeChanged",
                ((self.beam_info_dict["size_x"], self.beam_info_dict["size_y"]),),
            )
            self.emit("beamInfoChanged", (self.beam_info_dict,))


def test_hwo(hwo):
    print(hwo.get_beam_info())
    print(hwo.get_beam_position())



import logging
import math
import re

import gevent
from tango import DeviceProxy

from mxcubecore.HardwareObjects.abstract.AbstractMachineInfo import AbstractMachineInfo
from mxcubecore.utils.units import (
    A_to_mA,
    sec_to_hour,
)

# how often we refresh machine info
REFRESH_PERIOD_SEC = 30
CLEANR = re.compile("<.*?>")


log = logging.getLogger("HWR")


def cleanhtml(raw_html):
    cleantext = re.sub(CLEANR, "", raw_html)
    return cleantext


def catch_errors(func):
    """
    run wrapped function, catching all exception

    If an exception as raised, log the exception and return 'unknown'
    """

    def wrapper(*a, **kw):
        try:
            return func(*a, **kw)
        except Exception:
            log.exception("error fetching machine info")
            return "unknown"

    return wrapper


class MachInfo(AbstractMachineInfo):
    """
    Machine info hardware object for MAXIV site.

    Provides to the user general information about the machine,
    such as ring status, current, operator message, etc.

    This hardware objects fetches the information from attributes
    of specified tango devices.

    Hardware object properties:
        mach_info (str): name of the machine status tango device
        current (str): name of the ring status tango device
        parameters (str): topics to export, see AbstractMachineInfo class for details
    """

    def __init__(self, name):
        super().__init__(name)
        self.mach_info = None
        self.mach_curr = None

    def init(self):
        super().init()

        self.mach_info = self._get_tango_device("mach_info")
        self.mach_curr = self._get_tango_device("current")
        gevent.spawn(self._refresh_ticker)

    def _get_tango_device(self, property_name: str) -> DeviceProxy:
        dev_name = self.get_property(property_name)
        try:
            return DeviceProxy(dev_name)
        except Exception:
            log.exception(f"error connecting to machine info tango device {dev_name}")

    def _refresh_ticker(self):
        while True:
            self.update_value()
            gevent.sleep(REFRESH_PERIOD_SEC)

    @catch_errors
    def get_current(self) -> str:
        current = A_to_mA(self.mach_curr.Current)
        return f"{current:.2f} mA"

    @catch_errors
    def get_fillmode(self) -> str:
        return self.mach_info.R3Mode

    @catch_errors
    def get_message(self) -> str:
        return self.mach_info.OperatorMessage

    @catch_errors
    def get_lifetime(self) -> str:
        lifetime = self.mach_curr.Lifetime
        if math.isnan(lifetime):
            return "n/a"

        return f"{sec_to_hour(lifetime):.2f} h"

    @catch_errors
    def get_injection(self) -> str:
        return self.mach_info.R3NextInjection

    @catch_errors
    def get_status(self) -> str:
        message = cleanhtml(self.mach_info.MachineMessage)
        message = message.replace("R1", " R1").replace("Linac", " Linac")

        return message



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

""" """

import abc
import logging
from datetime import datetime
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model.lims_session import (
    Lims,
    LimsSessionManager,
    LimsUser,
    Session,
)

__credits__ = ["MXCuBE collaboration"]

StoreEvent = Literal["CREATE", "UPDATE", "END"]


class AbstractLims(HardwareObject, abc.ABC):
    """Interface for LIMS integration"""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)

        # current lims session
        self.active_session = None

        self.beamline_name = "unknown"

        self.sessions = []

        self.session_manager = LimsSessionManager()

    def is_session_already_active(self, session_id: str) -> bool:
        # If current selected session is already selected no need to do
        # anything else
        active_session = self.session_manager.active_session
        if active_session is not None:
            if active_session.session_id == session_id:
                return True
        return False

    @abc.abstractmethod
    def get_lims_name(self) -> List[Lims]:
        """
        Returns the LIMS used, name and description
        """
        raise Exception("Abstract class. Not implemented")

    def get_session_id(self) -> str:
        """
        Returns the currently active LIMS session id
        """
        return self.session_manager.active_session.session_id

    @abc.abstractmethod
    def get_user_name(self) -> str:
        """
        Returns the user name of the current user
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def get_full_user_name(self) -> str:
        """
        Returns the user name of the current user
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def login(
        self, login_id: str, password: str, create_session: bool
    ) -> List[Session]:
        """
        Login to LIMS, returns a list of Session objects for login_id

        Args:
            login_id: username
            password: password
            create_session: True if session should be created by LIMS if it
                            does not exist otherwise False

        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def is_user_login_type(self) -> bool:
        """
        Returns True if the login type is user based (not done with proposal)
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def echo(self) -> bool:
        """
        Returns True of LIMS is responding
        """
        raise Exception("Abstract class. Not implemented")

    def init(self) -> None:
        """
        Method inherited from baseclass
        """
        self.beamline_name = HWR.beamline.session.beamline_name

    @abc.abstractmethod
    def get_proposals_by_user(self, login_id: str) -> List[Dict]:
        """
        Returns a list with proposal dictionaries for login_id

        Proposal dictionary structure:
            {
                "Proposal": proposal,
                "Person": ,
                "Laboratory":,
                "Session":,
            }

        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def create_session(self, proposal_tuple: LimsSessionManager) -> LimsSessionManager:
        """
        TBD
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def get_samples(self, lims_name: str) -> List[Dict]:
        """
        Returns a list of sample dictionaries for the current user from lims_name

        Structure of sample dictionary:
        {
            "containerCode": str,
            "containerSampleChangerLocation": int,
            "crystalId": int,
            "crystalSpaceGroup": str,
            "diffractionPlan": {
                "diffractionPlanId": int
             },
            "proteinAcronym": "str,
            "sampleId": int,
            "sampleLocation": int,
            "sampleName": str
        }
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def store_robot_action(self, robot_action_dict: dict):
        """
        Stores the robot action dictionary.

        Structure of robot_action_dictionary:
        {
            "actionType":str,
            "containerLocation": str,
            "dewarLocation":str,
            "message":str,
            "sampleBarcode":str,
            "sessionId":int,
            "sampleId":int.
            "startTime":str,
            "endTime":str,
            "xtalSnapshotAfter:str",
            "xtalSnapshotBefore:str",
        }

        Args:
            robot_action_dict: robot action dictionary as defined above
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def store_beamline_setup(self, session_id: str, bl_config_dict: dict) -> int:
        """
        Stores the beamline setup dict bl_config_dict for session_id

        Args:
            session_id: The session id that the beamline_setup should be
                        associated with.

            bl_config_dict: The dictionary with beamline settings.

        Returns:
            The id of the beamline setup.
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def store_image(self, image_dict: dict) -> None:
        """
        Stores (image parameters) <image_dict>

        Args:
            image_dict: A dictionary with image pramaters.
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def store_energy_scan(self, energyscan_dict: dict) -> None:
        """
        Store energyscan data

        Args:
            energyscan_dict: Energyscan data to store.

        Returns:
            Dictionary with the energy scan id {"energyScanId": int}
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def store_xfe_spectrum(self, xfespectrum_dict: dict):
        """
        Stores a XFE spectrum.

        Args:
            xfespectrum_dict: XFE scan data to store.

        Returns:
            Dictionary with the XFE scan id {"xfeFluorescenceSpectrumId": int}
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def store_workflow(self, workflow_dict: dict) -> Tuple[int, int, int]:
        """
        Stores worklflow data workflow_dict

        Structure of workflow_dict:
        {
            "workflow_id": int,
            "workflow_type": str,
            "comments": str,
            "log_file_path": str,
            "result_file_path": str,
            "status": str,
            "title": str,
            "grid_info_id": int,
            "dx_mm": float,
            "dy_mm": float,
            "mesh_angle": float,
            "steps_x": float,
            "steps_y": float,
            "xOffset": float,
            "yOffset": float,
        }

        Args:
            workflow_dict: worklflow data on the format above

        Returns:
            Tuple of ints workflow_id, workflow_mesh_id, grid_info_id
        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def store_data_collection(
        self,
        datacollection_dict: dict,
        beamline_config_dict: Optional[Dict],
    ) -> Tuple[int, int]:
        """
        Stores a datacollection, datacollection_dict, and beamline configuration, beamline_config_dict, at the time of collection

        Structure of datacollection_dict:
        {
            "oscillation_sequence":[{}
                "start": float,
                "range": float,
                "overlap": float,
                "number_of_images": float,
                "start_image_number": float
                "exposure_time", float,
                "kappaStart": float,
                "phiStart": float,
            }],
            "fileinfo:{
                "directory": str,
                "prefix": str
                "suffix": str,
                "template: str,
                "run_number" int
            }
            "status": str,
            "collection_id": int,
            "wavelength": float,
            "resolution":{
                "lower": float,
                "upper": float
            },
            "resolutionAtCorner": float,
            "detectorDistance": float
            "xBeam": float,
            "yBeam": float,
            "beamSizeAtSampleX": float
            "beamSizeAtSampleY": float,
            "beamShape": str,
            "slitGapHorizontal": float,
            "slitGapVertical": float,
            "synchrotronMode", float,
            "flux": float,
            "flux_end": float,
            "transmission" float,
            "undulatorGap1": float
            "undulatorGap2": float
            "undulatorGap3": float
            "xtalSnapshotFullPath1": str,
            "xtalSnapshotFullPath2": str,
            "xtalSnapshotFullPath3": str,
            "xtalSnapshotFullPath4": str,
            "centeringMethod": str,
            "actualCenteringPosition" str
            "group_id: int,
            "detector_id": int,
            "screening_sub_wedge_id": int,
            "collection_start_time": str #"%Y-%m-%d %H:%M:%S"
        }

        Structure of beamline_config_dict:
        {
            "synchrotron_name":str,
            "directory_prefix":str,
            "default_exposure_time":str,
            "minimum_exposure_time":str,
            "detector_fileext":str,
            "detector_type":str,
            "detector_manufacturer":str,
            "detector_binning_mode":str,
            "detector_model":str,
            "detector_px":int,
            "detector_py":int,
            "undulators":str,
            "focusing_optic":str,
            "monochromator_type":str,
            "beam_divergence_vertical":float,
            "beam_divergence_horizontal":float,
            "polarisation":float,
            "maximum_phi_speed":float,
            "minimum_phi_oscillation":float,
            "input_files_server":str,
        }

        Args:
            datacollection_dict: As defined above
            beamline_config_dict: As defined above

        Returns:
           Tuple data_collection_id, detector_id


        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def update_data_collection(
        self,
        datacollection_dict: dict,
    ) -> Tuple[int, int]:
        """
        Updates the collection with "collection_id", provided in datacollection_dict.

        Structure of datacollection_dict as defined in store_data_collection above.

        Args:
            datacollection_dict:

        """
        raise Exception("Abstract class. Not implemented")

    @abc.abstractmethod
    def finalize_data_collection(
        self,
        datacollection_dict: dict,
    ) -> Tuple[int, int]:
        """
        Finalizes the collection with "collection_id", provided in datacollection_dict.

        Structure of datacollection_dict as defined in store_data_collection above.

        Args:
            datacollection_dict:
        """
        raise Exception("Abstract class. Not implemented")

    def is_scheduled_on_host_beamline(self, beamline: str) -> bool:
        """
        TBD
        """
        return beamline.strip().upper() == self.override_beamline_name.strip().upper()

    def is_scheduled_now(self, start_date: str, end_date: str) -> bool:
        """
        TBD
        """
        return self.is_time_between(start_date, end_date)

    def is_time_between(self, start_date: str, end_date: str, check_time=None):
        """
        TBD
        """
        if start_date is None or end_date is None:
            return False

        begin_time = datetime.fromisoformat(start_date).date()
        end_time = datetime.fromisoformat(end_date).date()

        # If check time is not given, default to current UTC time
        check_time = check_time or datetime.utcnow().date()
        if begin_time <= check_time <= end_time:
            return True
        else:
            return False

    def __set_sessions(self, sessions: List[Session]):
        """
        Sets the current lims session
        :param session: lims session value
        :return:
        """
        logging.getLogger("HWR").debug(
            "%s sessions available for users %s"
            % (len(sessions), self.session_manager.users.keys())
        )
        self.session_manager.sessions = sessions
        self.emit("sessionsChanged", (sessions,))

    def get_active_session(self) -> Session:
        """
        Returns Currently active session
        """
        return self.session_manager.active_session

    def set_active_session_by_id(self, session_id: str) -> Session:
        """
        Sets session with session_id to active session

        Args:
            session_id: session id
        """
        raise Exception("Abstract class. Not implemented")

    def get_shared_sessions(self):
        # Step 1: Collect all session_ids for each user
        session_ids_by_user = {}

        # Step 2: Iterate over users and collect session ids
        for user_name, user in self.session_manager.users.items():
            session_ids_by_user[user_name] = {
                session.session_id for session in user.sessions
            }

        # Step 3: Find the intersection of session_ids (sessions shared by all users)
        if not session_ids_by_user:
            return []  # If no users, return empty list

        # Find the common session ids across all users
        common_session_ids = set.intersection(*session_ids_by_user.values())

        # Step 4: Retrieve the sessions with these common session_ids and ensure uniqueness
        shared_sessions = {}
        for user_name, user in self.session_manager.users.items():
            for session in user.sessions:
                if session.session_id in common_session_ids:
                    # Use session_id as the key to ensure uniqueness
                    shared_sessions[session.session_id] = session

        # Convert the dictionary values (which are unique) into a list
        return list(shared_sessions.values())

    def remove_user(self, user_name: str):
        if user_name in self.session_manager.users:
            del self.session_manager.users[user_name]
            logging.getLogger("HWR").debug("User %s has been removed" % user_name)
            self.__set_sessions(self.get_shared_sessions())

    def add_user_and_shared_sessions(self, user_name: str, sessions: List[Session]):
        """
        Stores the username and the shared sessions in the session manager object.
        The shared sessions represent the intersection of all sessions
        for each user currently connected.
        """
        self.session_manager.users[user_name] = LimsUser(
            user_name=user_name, sessions=sessions
        )
        logging.getLogger("HWR").debug(
            "User added to session manager, user_name=%s sessions=%s"
            % (user_name, len(sessions))
        )

        self.__set_sessions(self.get_shared_sessions())



# -*- coding: utf-8 -*-

"""Transmission control for MANACA.
Attenuator: ABS-300-12-C-DN-DN40 Flanges.
Foils: #1, 8 um Al;
       #2, 10 um Al;
       #3, 20 um Al;
       #4, 80 um Al;
       #5, 160 um Al;
       #6, 320 um Al;
       #7, 800 um Al;
       #8, 1500 um Al;
       #9, 8 um Ti;
       #10, 10 um Cu
       #11, 5 um Au;
       #12, 25 um Zr.
 Here we use the Beer-Lambert law to calculate the transmission/attenuation: I = I0 * e^-(MU * d)
 were I is the transmitted (final) beam intensity, I0 is the incident (initial) beam intensity,
 MU is the linear attenuation coefficient (cm^-1) and d is the thickness of the foil (cm).
"""

import argparse
import math

from epics import PV


def read_input():
    # Read inputs from terminal
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-e",
        "--energy",
        required=True,
        type=float,
        help="Energy of the x-ray beam [keV]",
    )
    args = vars(ap.parse_args())
    return args


def read_transmission(energy):
    # Expressions (interpolation) for calculations of MU (linear attenuation coefficient [cm^-1]) for each material in function of the beam energy.
    # The data used to obtain the expressions (curve fitting) are from  "https://physics.nist.gov/PhysRefData/FFast/html/form.html"
    # This attenuation coefficients can be verified and optimised after experimental validation at the beamline.
    MU_Al = (
        (78657.01011 * math.exp(-energy / 0.65969))
        + (6406.36151 * math.exp(-energy / 1.63268))
        + (492.29999 * math.exp(-energy / 4.42554))
        + (3.2588)
    )
    MU_Ti = (
        (28062.17632 * math.exp(-energy / 1.88062))
        + (6354120 * math.exp(-energy / 0.49973))
        + (2257.91488 * math.exp(-energy / 5.28296))
        + (17.4342)
    )
    MU_Cu = (
        (1147850000 * math.exp(-energy / 0.56933))
        + (2582.7593 * math.exp(-energy / 8.40671))
        + (30628.08291 * math.exp(-energy / 2.98937))
        + (22.4187)
    )
    MU_Au = (42783.04258 * math.exp(-energy / 5.39244)) + (399.59714)
    MU_Zr = (11143.28458 * math.exp(-energy / 5.87029)) + (102.27474)
    MU_EMPTY = 0.0

    foils_pv = {
        "F1_Al": (8, MU_Al, PV("MNC:B:RIO01:9425A:bi11"), PV("MNC:B:RIO01:9425A:bi12")),
        "F2_Al": (
            10,
            MU_Al,
            PV("MNC:B:RIO01:9425A:bi10"),
            PV("MNC:B:RIO01:9425A:bi13"),
        ),
        "F3_Al": (20, MU_Al, PV("MNC:B:RIO01:9425A:bi9"), PV("MNC:B:RIO01:9425A:bi14")),
        "F4_Al": (80, MU_Al, PV("MNC:B:RIO01:9425A:bi8"), PV("MNC:B:RIO01:9425A:bi15")),
        "F5_Al": (
            160,
            MU_Al,
            PV("MNC:B:RIO01:9425A:bi7"),
            PV("MNC:B:RIO01:9425A:bi16"),
        ),
        "F6_Al": (
            320,
            MU_Al,
            PV("MNC:B:RIO01:9425A:bi6"),
            PV("MNC:B:RIO01:9425A:bi17"),
        ),
        "F7_Al": (
            800,
            MU_Al,
            PV("MNC:B:RIO01:9425A:bi5"),
            PV("MNC:B:RIO01:9425A:bi18"),
        ),
        "F8_Al": (
            1500,
            MU_Al,
            PV("MNC:B:RIO01:9425A:bi4"),
            PV("MNC:B:RIO01:9425A:bi19"),
        ),
        "F9_Ti": (8, MU_Ti, PV("MNC:B:RIO01:9425A:bi3"), PV("MNC:B:RIO01:9425A:bi20")),
        "F10_Cu": (
            10,
            MU_Cu,
            PV("MNC:B:RIO01:9425A:bi2"),
            PV("MNC:B:RIO01:9425A:bi21"),
        ),
        "F11_Au": (5, MU_Au, PV("MNC:B:RIO01:9425A:bi1"), PV("MNC:B:RIO01:9425A:bi22")),
        "F12_Zr": (
            25,
            MU_Zr,
            PV("MNC:B:RIO01:9425A:bi0"),
            PV("MNC:B:RIO01:9425A:bi23"),
        ),
    }

    # check what foils are in the the beam and calculate MU*x and then calculate transmission

    # the commented lines below works for the attenuator expected logic (0 for foil out and 1 for foil in).
    # as the current logic is inverted (1 is foil out and 0 is foil in) use the uncommented lines.
    foils_in = []
    for foil in foils_pv.keys():
        # if foils_pv[foil][2].get() == 1 and foils_pv[foil][3].get() == 0:
        if foils_pv[foil][2].get() == 0 and foils_pv[foil][3].get() == 1:
            atten_mu_x = float(foils_pv[foil][0] * 1.0e-4) * foils_pv[foil][1]
            foils_in.append(atten_mu_x)
            status = 0
        # elif foils_pv[foil][2].get() == 0 and foils_pv[foil][3].get() == 1:
        elif foils_pv[foil][2].get() == 1 and foils_pv[foil][3].get() == 0:
            status = 0
        else:
            status = 1

    MU_x = sum(foils_in)
    transmission = 1.0 * (math.exp(-MU_x))

    return transmission, status


def main():
    energy = read_input()["energy"]
    T = read_transmission(energy)
    transmission = round(T[0] * 100, 2)
    foil_status = T[1]


if __name__ == "__main__":
    main()



import time

from mxcubecore.HardwareObjects.abstract.AbstractSampleChanger import (
    SC3,
    Container,
    Sample,
    SampleChanger,
    SampleChangerState,
    gevent,
)


class Pin(Sample):
    STD_HOLDERLENGTH = 22.0

    def __init__(self, basket, basket_no, sample_no):
        super(Pin, self).__init__(
            basket, Pin.get_sample_address(basket_no, sample_no), True
        )
        self._set_holder_length(Pin.STD_HOLDERLENGTH)

    def get_basket_no(self):
        return self.get_container().get_index() + 1

    def get_vial_no(self):
        return self.get_index() + 1

    @staticmethod
    def get_sample_address(basket_number, sample_number):
        return str(basket_number) + ":" + "%02d" % (sample_number)


class Basket(Container):
    __TYPE__ = "Puck"
    NO_OF_SAMPLES_PER_PUCK = 10

    def __init__(self, container, number):
        super(Basket, self).__init__(
            self.__TYPE__, container, Basket.get_basket_address(number), True
        )
        for i in range(Basket.NO_OF_SAMPLES_PER_PUCK):
            slot = Pin(self, number, i + 1)
            self._add_component(slot)

    @staticmethod
    def get_basket_address(basket_number):
        return str(basket_number)

    def clear_info(self):
        self.get_container()._reset_basket_info(self.get_index() + 1)
        self.get_container()._trigger_info_changed_event()


class CatsBessy(SampleChanger):
    __TYPE__ = "CATS"
    NO_OF_BASKETS = 9

    """
    Actual implementation of the CATS Sample Changer, BESSY BL14.1 installation
    """

    def __init__(self, *args, **kwargs):
        super(CatsBessy, self).__init__(self.__TYPE__, False, *args, **kwargs)
        for i in range(CatsBessy.NO_OF_BASKETS):
            basket = Basket(self, i + 1)
            self._add_component(basket)

    def init(self):
        self._selected_sample = 1
        self._selected_basket = 1

        self._state = self.get_channel_object("_state")
        self._abort = self.get_command_object("_abort")

        self._basketChannels = []
        for basket_index in range(CatsBessy.NO_OF_BASKETS):
            self._basketChannels.append(
                self.add_channel(
                    {
                        "type": "tango",
                        "name": "di_basket",
                        "tangoname": self.tangoname,
                        "polling": "events",
                    },
                    ("di_Cassette%dPresence" % (basket_index + 1)),
                )
            )

        self._lidStatus = self.add_channel(
            {
                "type": "tango",
                "name": "di_AllLidsClosed",
                "tangoname": self.tangoname,
                "polling": "events",
            },
            "di_AllLidsClosed",
        )
        if self._lidStatus is not None:
            self._lidStatus.connect_signal("update", self._update_operation_mode)
        self._scIsCharging = None

        self._load = self.add_command(
            {"type": "tango", "name": "put_bcrd", "tangoname": self.tangoname},
            "put_bcrd",
        )
        self._unload = self.add_command(
            {"type": "tango", "name": "put_bcrd", "tangoname": self.tangoname}, "get"
        )
        self._chained_load = self.add_command(
            {"type": "tango", "name": "getput_bcrd", "tangoname": self.tangoname},
            "getput_bcrd",
        )
        self._barcode = self.add_command(
            {"type": "tango", "name": "barcode", "tangoname": self.tangoname}, "barcode"
        )
        self._reset = self.add_command(
            {"type": "tango", "name": "reset", "tangoname": self.tangoname}, "reset"
        )
        self._abort = self.add_command(
            {"type": "tango", "name": "abort", "tangoname": self.tangoname}, "abort"
        )

        self._numSampleOnDiff = self.add_channel(
            {
                "type": "tango",
                "name": "NumSampleOnDiff",
                "tangoname": self.tangoname,
                "polling": "events",
            },
            "NumSampleOnDiff",
        )
        self._lidSampleOnDiff = self.add_channel(
            {
                "type": "tango",
                "name": "LidSampleOnDiff",
                "tangoname": self.tangoname,
                "polling": "events",
            },
            "LidSampleOnDiff",
        )
        self._barcode = self.add_channel(
            {
                "type": "tango",
                "name": "Barcode",
                "tangoname": self.tangoname,
                "polling": "events",
            },
            "Barcode",
        )
        self._path_running = self.add_channel(
            {
                "type": "tango",
                "name": "PathRunning",
                "tangoname": self.tangoname,
                "polling": "events",
            },
            "PathRunning",
        )

        self._init_sc_contents()

        # SampleChanger.init must be called _after_ initialization of the Cats because it starts the update methods which access
        # the device server's status attributes
        SampleChanger.init(self)

    def get_sample_properties(self):
        return (Pin.__HOLDER_LENGTH_PROPERTY__,)

    # ########################           TASKS           #########################
    def _do_update_info(self):
        self._updateSCContents()
        self._update_selection()
        self._update_state()
        self._update_loaded_sample()

    def _do_change_mode(self, mode):
        pass

    def get_selected_component(self):
        return self.get_components()[self._selected_basket - 1]

    def _do_select(self, component):
        if isinstance(component, Sample):
            selected_basket = self.get_selected_component()
            if (selected_basket is None) or (
                selected_basket != component.get_container()
            ):
                # self._execute_server_task(self._select_basket , component.get_basket_no())
                self._selected_basket = component.get_basket_no()
            # self._execute_server_task(self._select_sample, component.get_index()+1)
            self._selected_sample = component.get_index() + 1
        elif isinstance(component, Container) and (
            component.get_type() == Basket.__TYPE__
        ):
            # self._execute_server_task(self._select_basket, component.get_index()+1)
            self._selected_basket = component.get_index() + 1

    def _do_abort(self):
        self._abort()

    def _do_scan(self, component, recursive):
        selected_basket = self.get_selected_component()
        if isinstance(component, Sample):
            # scan a single sample
            if (selected_basket is None) or (
                selected_basket != component.get_container()
            ):
                self._do_select(component)
            # self._execute_server_task(self._scan_samples, [component.get_index()+1,])
            lid = ((self._selected_basket - 1) / 3) + 1
            sample = (((self._selected_basket - 1) % 3) * 10) + (
                component.get_index() + 1
            )
            argin = ["2", str(lid), str(sample), "0", "0"]
            self._execute_server_task(self._barcode, argin)
        elif isinstance(component, Container) and (
            component.get_type() == Basket.__TYPE__
        ):
            # component is a basket
            if recursive:
                pass
                # self._execute_server_task(self._scan_basket, (component.get_index()+1))
            else:
                if (selected_basket is None) or (selected_basket != component):
                    self._do_select(component)
                # self._execute_server_task(self._scan_samples, (0,))
                for sample_index in range(Basket.NO_OF_SAMPLES_PER_PUCK):
                    lid = ((self._selected_basket - 1) / 3) + 1
                    sample = (((self._selected_basket - 1) % 3) * 10) + (
                        sample_index + 1
                    )
                    argin = ["2", str(lid), str(sample), "0", "0"]
                    self._execute_server_task(self._barcode, argin)
        elif isinstance(component, Container) and (
            component.get_type() == SC3.__TYPE__
        ):
            for basket in self.get_components():
                self._do_scan(basket, True)

    def _do_load(self, sample=None):
        selected = self.get_selected_sample()
        if self.has_loaded_sample():
            if (sample is None) or (sample == self.get_loaded_sample()):
                raise Exception(
                    "The sample "
                    + str(self.get_loaded_sample().get_address())
                    + " is already loaded"
                )
            lid = ((self._selected_basket - 1) / 3) + 1
            sample = (((self._selected_basket - 1) % 3) * 10) + self._selected_sample
            argin = ["2", str(lid), str(sample), "0", "0", "0", "0", "0"]
            self._execute_server_task(self._chained_load, argin)
        else:
            if sample is None:
                if selected is None:
                    raise Exception("No sample selected")
                else:
                    sample = selected
            elif (sample is not None) and (sample != selected):
                self._do_select(sample)
            # self._execute_server_task(self._load,sample.get_holder_length())
            lid = ((self._selected_basket - 1) / 3) + 1
            sample = (((self._selected_basket - 1) % 3) * 10) + self._selected_sample
            argin = ["2", str(lid), str(sample), "0", "0", "0", "0", "0"]
            self._execute_server_task(self._load, argin)

    def _do_unload(self, sample_slot=None):
        if sample_slot is not None:
            self._do_select(sample_slot)
        argin = ["2", "0", "0", "0", "0"]
        self._execute_server_task(self._unload, argin)

    def _do_reset(self):
        self._execute_server_task(self._reset)

    def clear_basket_info(self, basket):
        self._reset_basket_info(basket)

    # ########################           PRIVATE           #########################
    def _update_operation_mode(self, value):
        self._scIsCharging = not value

    def _execute_server_task(self, method, *args):
        self._wait_device_ready(3.0)
        task_id = method(*args)
        # introduced wait because it takes some time before the attribute PathRunning is set
        # after launching a transfer
        time.sleep(2.0)
        ret = None
        if task_id is None:  # Reset
            while self._is_device_busy():
                gevent.sleep(0.1)
        else:
            while str(self._path_running.get_value()).lower() == "true":
                gevent.sleep(0.1)
            # try:
            #    ret = self._check_task_result(task_id)
            # except Exception,err:
            #    raise
            ret = True
        return ret

    def _update_state(self):
        try:
            state = self._read_state()
        except Exception:
            state = SampleChangerState.Unknown
        if state == SampleChangerState.Moving and self._is_device_busy(
            self.get_state()
        ):
            return
        if self._scIsCharging and not (state == SampleChangerState.Alarm):
            state = SampleChangerState.Charging
        self._set_state(state)

    def _read_state(self):
        state = self._state.get_value()
        if state is not None:
            stateStr = str(state).upper()
        else:
            stateStr = ""
        # state = str(self._state.get_value() or "").upper()
        state_converter = {
            "ALARM": SampleChangerState.Alarm,
            "ON": SampleChangerState.Ready,
            "RUNNING": SampleChangerState.Moving,
        }
        return state_converter.get(stateStr, SampleChangerState.Unknown)

    def _is_device_busy(self, state=None):
        if state is None:
            state = self._read_state()
        return state not in (
            SampleChangerState.Ready,
            SampleChangerState.Loaded,
            SampleChangerState.Alarm,
            SampleChangerState.Disabled,
            SampleChangerState.Fault,
            SampleChangerState.StandBy,
        )

    def _is_device_ready(self):
        state = self._read_state()
        return state in (SampleChangerState.Ready, SampleChangerState.Charging)

    def _wait_device_ready(self, timeout=None):
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while not self._is_device_ready():
                gevent.sleep(0.01)

    def _update_selection(self):
        basket = None
        sample = None
        try:
            basket_no = self._selected_basket
            if (
                basket_no is not None
                and basket_no > 0
                and basket_no <= CatsBessy.NO_OF_BASKETS
            ):
                basket = self.get_component_by_address(
                    Basket.get_basket_address(basket_no)
                )
                sample_no = self._selected_sample
                if (
                    sample_no is not None
                    and sample_no > 0
                    and sample_no <= Basket.NO_OF_SAMPLES_PER_PUCK
                ):
                    sample = self.get_component_by_address(
                        Pin.get_sample_address(basket_no, sample_no)
                    )
        except Exception:
            pass
        self._set_selected_component(basket)
        self._set_selected_sample(sample)

    def _update_loaded_sample(self):
        loadedSampleLid = self._lidSampleOnDiff.get_value()
        loadedSampleNum = self._numSampleOnDiff.get_value()
        if loadedSampleLid != -1 or loadedSampleNum != -1:
            lidBase = (loadedSampleLid - 1) * 3
            lidOffset = ((loadedSampleNum - 1) / 10) + 1
            samplePos = ((loadedSampleNum - 1) % 10) + 1
            basket = lidBase + lidOffset
        else:
            basket = None
            samplePos = None

        if basket is not None and samplePos is not None:
            new_sample = self.get_component_by_address(
                Pin.get_sample_address(basket, samplePos)
            )
        else:
            new_sample = None

        if self.get_loaded_sample() != new_sample:
            # remove 'loaded' flag from old sample but keep all other information
            old_sample = self.get_loaded_sample()
            if old_sample is not None:
                # there was a sample on the gonio
                loaded = False
                has_been_loaded = True
                old_sample._set_loaded(loaded, has_been_loaded)
            if new_sample is not None:
                # update information of recently loaded sample
                datamatrix = str(self._barcode.get_value())
                scanned = len(datamatrix) != 0
                if not scanned:
                    datamatrix = "----------"
                loaded = True
                has_been_loaded = True
                new_sample._set_info(new_sample.is_present(), datamatrix, scanned)
                new_sample._set_loaded(loaded, has_been_loaded)

    def _init_sc_contents(self):
        # create temporary list with default basket information
        basket_list = [("", 4)] * CatsBessy.NO_OF_BASKETS
        # write the default basket information into permanent Basket objects
        for basket_index in range(CatsBessy.NO_OF_BASKETS):
            basket = self.get_components()[basket_index]
            datamatrix = None
            present = scanned = False
            basket._set_info(present, datamatrix, scanned)

        # create temporary list with default sample information and indices
        sample_list = []
        for basket_index in range(CatsBessy.NO_OF_BASKETS):
            for sample_index in range(Basket.NO_OF_SAMPLES_PER_PUCK):
                sample_list.append(
                    ("", basket_index + 1, sample_index + 1, 1, Pin.STD_HOLDERLENGTH)
                )
        # write the default sample information into permanent Pin objects
        for spl in sample_list:
            sample = self.get_component_by_address(
                Pin.get_sample_address(spl[1], spl[2])
            )
            datamatrix = None
            present = scanned = loaded = has_been_loaded = False
            sample._set_info(present, datamatrix, scanned)
            sample._set_loaded(loaded, has_been_loaded)
            sample._set_holder_length(spl[4])

    def _updateSCContents(self):
        for basket_index in range(CatsBessy.NO_OF_BASKETS):
            # get presence information from the device server
            newBasketPresence = self._basketChannels[basket_index].get_value()
            # get saved presence information from object's internal bookkeeping
            basket = self.get_components()[basket_index]

            # check if the basket was newly mounted or removed from the dewar
            if newBasketPresence ^ basket.is_present():
                # a mounting action was detected ...
                if newBasketPresence:
                    # basket was mounted
                    present = True
                    scanned = False
                    datamatrix = None
                    basket._set_info(present, datamatrix, scanned)
                else:
                    # basket was removed
                    present = False
                    scanned = False
                    datamatrix = None
                    basket._set_info(present, datamatrix, scanned)
                # set the information for all dependent samples
                for sample_index in range(Basket.NO_OF_SAMPLES_PER_PUCK):
                    sample = self.get_component_by_address(
                        Pin.get_sample_address((basket_index + 1), (sample_index + 1))
                    )
                    present = sample.get_container().is_present()
                    if present:
                        datamatrix = "          "
                    else:
                        datamatrix = None
                    scanned = False
                    sample._set_info(present, datamatrix, scanned)
                    # forget about any loaded state in newly mounted or removed basket)
                    loaded = has_been_loaded = False
                    sample._set_loaded(loaded, has_been_loaded)



# -*- coding: utf-8 -*-
import logging

from PyTango import DeviceProxy

from mxcubecore.BaseHardwareObjects import HardwareObject

"""Complex means we are not using SimpleDevice"""


class SOLEILPss(HardwareObject):
    states = {0: "not ready", 1: "ready"}

    READ_CMD, READ_OUT = (0, 1)

    def __init__(self, name):
        super().__init__(name)

        self.wagoState = "unknown"
        self.__oldValue = None
        self.device = None
        self.hutch = None

    def init(self):
        try:
            self.device = DeviceProxy(self.get_property("tangoname"))
        except Exception:
            logging.getLogger("HWR").error(
                "%s: unknown pss device name", self.get_property("tangoname")
            )

        if self.get_property("hutch") not in ("optical", "experimental"):
            logging.getLogger("HWR").error(
                "SOLEILPss.init Hutch property %s is not correct",
                self.get_property("hutch"),
            )
        else:
            self.hutch = self.get_property("hutch")
            self.stateChan = self.get_channel_object("State")
            self.stateChan.connect_signal("update", self.value_changed)
        if self.device:
            self.set_is_ready(True)

    def get_state(self, value):
        return SOLEILPss.states[value]

    def getWagoState(self):
        return self.get_state(self.stateChan.get_value())

    def value_changed(self, value):
        logging.getLogger("HWR").info("%s: SOLEILPss.valueChanged, %s", self.id, value)
        state = self.get_state(value)
        self.emit("wagoStateChanged", (state,))



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.model import queue_model_objects
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class SampleCentringQueueEntry(BaseQueueEntry):
    """
    Entry for centring a sample
    """

    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __getstate__(self):
        d = dict(self.__dict__)
        d["move_kappa_phi_task"] = None
        return d

    def execute(self):
        BaseQueueEntry.execute(self)

        self.get_view().setText(1, "Waiting for input")
        log = logging.getLogger("user_level_log")

        data_model = self.get_data_model()

        kappa = data_model.get_kappa()
        kappa_phi = data_model.get_kappa_phi()

        # kappa and kappa_phi settings are applied first, and assume that the
        # beamline does have axes with exactly these names
        #
        # Other motor_positions are applied afterwards, but in random order.
        # motor_positions override kappa and kappa_phi if both are set
        #
        # Since setting one motor can change the position of another
        # (on ESRF ID30B setting kappa and kappa_phi changes the translation motors)
        # the order is important.
        dd0 = {}
        if kappa is not None:
            dd0["kappa"] = kappa
        if kappa_phi is not None:
            dd0["kappa_phi"] = kappa_phi
        if dd0:
            if (
                not hasattr(HWR.beamline.diffractometer, "in_kappa_mode")
                or HWR.beamline.diffractometer.in_kappa_mode()
            ):
                HWR.beamline.diffractometer.move_motors(dd0)

        motor_positions = dict(
            tt0
            for tt0 in data_model.get_other_motor_positions().items()
            if tt0[1] is not None
        )
        if motor_positions:
            HWR.beamline.diffractometer.move_motors(motor_positions)

        log.warning("Please center a new or select an existing point and press resume.")
        self.get_queue_controller().pause(True)

        shapes = list(HWR.beamline.sample_view.get_selected_shapes())

        if shapes:
            pos = shapes[0]
            if hasattr(pos, "get_centred_position"):
                cpos = pos.get_centred_position()
            else:
                cpos = pos.get_centred_positions()[0]
        else:
            msg = "No centred position selected, using current position."
            log.info(msg)

            # Create a centred positions of the current position
            pos_dict = HWR.beamline.diffractometer.get_positions()
            cpos = queue_model_objects.CentredPosition(pos_dict)

        self._data_model.set_centring_result(cpos)

        self.get_view().setText(1, "Input accepted")

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)

    def post_execute(self):
        # If centring is executed once then we dont have to execute it again
        self.get_view().set_checkable(False)
        BaseQueueEntry.post_execute(self)

    def get_type_str(self):
        return "Sample centering"



import logging

from devtools import debug

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.mx_base_queue_entry import (
    BaseUserCollectionParameters,
    MXBaseQueueEntry,
    MXBaseQueueTaskParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class FastCharUserCollectionParameters(BaseUserCollectionParameters):
    class Config:
        extra: "ignore"


class FastCharCollectionTaskParameters(MXBaseQueueTaskParameters):
    user_collection_parameters: FastCharUserCollectionParameters

    @staticmethod
    def update_dependent_fields(field_data):
        new_data = {}
        return new_data


class FastCharCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class FastCharCollectionQueueEntry(MXBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = FastCharCollectionQueueModel
    DATA_MODEL = FastCharCollectionTaskParameters
    NAME = "Fast characterisation"
    REQUIRES = ["point", "line", "no_shape", "chip", "mesh"]

    def __init__(self, view, data_model: FastCharCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()
        debug(self._data_model._task_data)
        data_path = self.get_data_path()

        number_of_images = 10
        number_of_scans = 4
        angle = 90
        exptime = 0.1
        osc_range = 1
        osc_start = 0

        HWR.beamline.diffractometer.set_phase("DataCollection")

        logging.getLogger("user_level_log").info(f"Preparing acqusisition")
        HWR.beamline.detector.prepare_acquisition(
            0,
            osc_start,
            osc_range,
            exptime / number_of_images,
            None,
            int(number_of_images * number_of_scans),
            "",
            True,
            0,
        )

        logging.getLogger("user_level_log").info(f"Setting file path to {data_path}")
        HWR.beamline.detector.set_detector_filenames(1, 1, data_path)

        HWR.beamline.diffractometer.wait_ready()
        logging.getLogger("user_level_log").info(f"Phase changed to data collection")

        HWR.beamline.detector.start_acquisition()
        logging.getLogger("user_level_log").info(f"Acquiring ...")
        HWR.beamline.diffractometer.characterisation_scan(
            osc_start,
            osc_range,
            number_of_images,
            exptime,
            number_of_scans,
            angle,
            wait=True,
        )

        HWR.beamline.diffractometer.wait_ready()
        HWR.beamline.detector.wait_ready()
        logging.getLogger("user_level_log").info(f"Finished fast characterisation")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



"""

Notes:

   EnergyScan and Energy are now separate hardware objects.

   Methods for Managing Energy and Wavelength are now removed from
   this hardware object (V.Rey - Jan 2018)

   This has been modified to follow the AbstractEnergyScan method

"""

import logging
import math
import os
import subprocess
import time

import gevent
import numpy
from AbstractEnergyScan import AbstractEnergyScan
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure
from xabs_lib import McMaster

from mxcubecore import HardwareRepository as HWR
from mxcubecore.Command.Tango import DeviceProxy
from mxcubecore.TaskUtils import (
    cleanup,
    task,
)


class PX1EnergyScan(AbstractEnergyScan, Equipment):
    round_cutoff = 4
    roi_width = 0.30

    default_steps = 80
    before_edge = 0.035
    after_edge = 0.045
    scan_range = after_edge + before_edge

    integration_time = 1
    thEdgeThreshold = 0.01

    def __init__(self, name):
        AbstractEnergyScan.__init__(self)
        super().__init__(name)

        self.scanning = False
        self.stopping = False

        self.ready_event = None

        self.e_edge = ""

        self.scan_info = {}
        self.scan_data = []

        self.log = logging.getLogger("HWR")

    def init(self):
        self.ready_event = gevent.event.Event()

        self.ruche_hwo = self.get_object_by_role("ruche")

        self.fluodet_hwo = self.get_object_by_role("fluodet")
        self.px1env_hwo = self.get_object_by_role("px1environment")

        self.mono_dp = DeviceProxy(self.get_property("mono_dev"))
        self.ble_dp = DeviceProxy(self.get_property("ble_dev"))
        self.fp_dp = DeviceProxy(self.get_property("fp_dev"))

        test_data_file = self.get_property("test_data")

        self.log.debug(" using test data %s" % test_data_file)

        self.test_data_mode = False

        if test_data_file:
            self.simul_data = self.load_data_file(test_data_file)
            if len(self.simul_data):
                self.test_data_mode = True

        # USING TEST. UNCOMMENT NEXT LINE TO USE REAL DATA IN ALL CASES
        # self.test_data_mode = False

        normdiode = self.get_property("normalization_diode")
        self.norm_diode_dev = DeviceProxy(normdiode)

        self.number_of_steps = self.get_property("number_of_steps")
        if self.number_of_steps is None:
            self.number_of_steps = self.default_steps

    def is_connected(self):
        return True

    # SCAN info (for graph)
    def new_data_point(self, x, y):
        logging.info("EnergyScan newPoint %s, %s" % (x, y))
        x = (x < 1000) and x * 1000 or x

        self.emit("scanNewPoint", (x, y))
        self.emit("progressStep", (len(self.scan_data)))

        self.scan_data.append((x, y))

    # SCAN info end (for graph)

    # Calculate scan values
    def get_edge_from_xabs(self, element, edge):
        edge = edge.upper()
        if edge[0] == "K":
            roi_center = McMaster[element]["edgeEnergies"]["K-alpha"]
        elif edge[0] == "L":
            roi_center = McMaster[element]["edgeEnergies"]["L-alpha"]
        if edge[0] == "L":
            edge = "L3"
        e_edge = McMaster[element]["edgeEnergies"][edge]

        return e_edge, roi_center

    def get_egy_values_to_scan(self):
        nb_steps = self.number_of_steps
        points = numpy.arange(0.0, 1.0 + 1.0 / (nb_steps), 1.0 / (nb_steps))

        points *= self.scan_range
        points -= self.before_edge
        points += self.e_edge
        points = numpy.array(map(self.round_egy, points))

        return points

    def round_egy(self, en):
        return round(en, self.round_cutoff)

    # Calculate scan values end

    # HARDWARE ACCESS
    def move_mono(self, energy):
        self.mono_dp.energy = float(energy)
        self.wait_device(self.mono_dp)

    def wait_device(self, device):
        while device.state().name in ["MOVING", "RUNNING"]:
            time.sleep(0.1)

    def go_to_collect(self, timeout=30):
        if self.px1env_hwo.isPhaseFluoScan():
            return

        self.px1env_hwo.gotoFluoScanPhase()

        t0 = time.time()
        while not self.px1env_hwo.isPhaseFluoScan():
            if (time.time() - t0) > timeout:
                logging.debug("PX1EnergyScan - Timed out while going to FluoXPhase")
                break
            time.sleep(0.1)

    def get_transmission(self):
        """Get or set the transmission"""
        return self.fp_dp.TrueTrans_FP

    def move_beamline_energy(self, energy):
        current_energy = self.ble_dp.energy
        if abs(current_energy - energy) > 0.001:
            self.ble_dp.energy = energy
            self.wait_device(self.ble_dp)

    def move_egy_to_peak(self, pk):
        self.move_beamline_energy(pk)

    def open_fast_shutter(self):
        HWR.beamline.fast_shutter.openShutter()

    def close_fast_shutter(self):
        HWR.beamline.fast_shutter.closeShutter()

    def close_safety_shutter(self):
        HWR.beamline.safety_shutter.closeShutter()

    def fluodet_prepare(self):
        self.fluodet_hwo.set_preset(float(self.integration_time))

    def set_mca_roi(self):
        # calibration
        A = -0.0161723871876
        B = 0.00993475667754
        C = 0.0

        # roi_center = A + B*self.roi_center + C*self.roi_center**2
        self.roi_center += A

        egy_start = 1000.0 * (self.roi_center - self.roi_width / 2.0)
        egy_end = 1000.0 * (self.roi_center + self.roi_width / 2.0)

        self.roi_start_chan = int(egy_start / (B * 1.0e3))
        self.roi_end_chan = int(egy_end / (B * 1e3))

        self.fluodet_hwo.set_roi(self.roi_start_chan, self.roi_end_chan)

    def acquire_point(self, en):
        self.open_fast_shutter()
        self.fluodet_hwo.start()
        self.fluodet_hwo.wait()
        self.close_fast_shutter()

        roi_counts = self.fluodet_hwo.get_roi_counts()
        intensity = self.norm_diode_dev.intensity

        point = float(roi_counts / intensity)

        return point

    def stop(self):
        self.stopping = True

    def abort(self):
        self.stop()
        self.close_fast_shutter()

    # HARDWARE ACCESS END

    # GENERAL SCAN PROGRAMMING
    def escan_prepare(self):
        self.stopping = False

        self.scan_data = []

        self.go_to_collect()

        self.e_edge, self.roi_center = self.get_edge_from_xabs(
            self.scan_info["element"], self.scan_info["edgeEnergy"]
        )

        self.set_mca_roi()
        self.fluodet_prepare()

        self.points = self.get_egy_values_to_scan()
        self.log.info(
            "   edge for %s/%s is %s"
            % (self.scan_info["element"], self.scan_info["edgeEnergy"], self.e_edge)
        )
        self.log.info("   points : %s" % str(self.points))

        self.num_points = len(self.points)
        self.ble_value = self.e_edge + 0.01

        self.scan_info["fluorescenceDetector"] = "Ketek detector"
        self.scan_info["transmissionFactor"] = self.get_transmission()
        self.scan_info["exposureTime"] = self.integration_time
        self.scan_info["startEnergy"] = 0.0  # updated at the end of the scan
        self.scan_info["endEnergy"] = 0.0  # updated at the end of the scan

        try:
            size_hor, size_ver = HWR.beamline.beam.get_beam_size()
            size_hor *= 1000
            size_ver *= 1000
        except Exception:
            size_hor = None
            size_ver = None

        self.scan_info["beamSizeHorizontal"] = size_hor
        self.scan_info["beamSizeVertical"] = size_ver

    def start_energy_scan(
        self, element, edge, directory, prefix, session_id=None, blsample_id=None
    ):
        log = logging.getLogger("HWR")

        self.scan_info = {
            "sessionId": session_id,
            "blSampleId": blsample_id,
            "element": element,
            "edgeEnergy": edge,
        }

        self.scan_data = []

        self.scan_prefix = prefix
        self.scan_directory = directory

        if not os.path.isdir(directory):
            log.debug("EnergyScan: creating directory %s" % directory)
            try:
                os.makedirs(directory)
            except OSError as diag:
                log.error(
                    "EnergyScan: error creating directory %s (%s)"
                    % (directory, str(diag))
                )
                self.emit("scanStatusChanged", ("Error creating directory",))
                return False

        try:
            self.escan_prepare()

            self.scanCommandStarted()

            self.move_beamline_energy(self.ble_value)

            for en in self.points:
                if self.stopping:
                    break

                self.move_mono(en)
                point_value = self.acquire_point(en)
                self.new_data_point(en, point_value)

            self.ready_event.set()

            self.scanCommandFinished()
        except Exception:
            import traceback

            logging.getLogger("HWR").error(
                "EnergyScan: problem starting energy scan. %s" % traceback.format_exc()
            )
            self.scanCommandFailed()
            self.scanStatusChanged("error starting energy scan")
            return False

        return True

    def cancelEnergyScan(self, *args):
        if self.scanning:
            self.abort()
            self.scanCommandAborted()

    def scanStatusChanged(self, status):
        self.emit("scanStatusChanged", (status,))

    def scanCommandStarted(self, *args):
        self.scan_info["startTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.scanning = True

        if self.scan_info["blSampleId"]:
            title = "Sample: %s Element: %s Edge: %s" % (
                self.scan_info["blSampleId"],
                self.scan_info["element"],
                self.scan_info["edgeEnergy"],
            )
        else:
            title = "Element: %s Edge: %s" % (
                self.scan_info["element"],
                self.scan_info["edgeEnergy"],
            )

        graph_info = {
            "xlabel": "energy",
            "ylabel": "counts",
            "scaletype": "normal",
            "title": title,
        }

        self.emit("energyScanStarted", graph_info)
        self.emit("progressInit", "Energy scan", self.num_points)

    def scanCommandFailed(self, *args):
        with cleanup(self.ready_event.set):
            self.scan_info["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")

            if self.scan_data:
                self.scan_info["startEnergy"] = self.scan_data[-1][0]
                self.scan_info["endEnergy"] = self.scan_data[-1][1]

            self.emit("energyScanFailed", ())
            self.emit("progressStop", ())

            self.scanning = False
            self.ready_event.set()

    def scanCommandAborted(self, *args):
        self.emit("energyScanFailed", ())
        self.emit("progressStop", ())
        self.scanning = False
        self.ready_event.set()

    def scanCommandFinished(self, *args):
        logging.getLogger("HWR").debug("EnergyScan: finished")

        with cleanup(self.ready_event.set):
            self.scan_info["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")

            self.scanning = False

            self.scan_info["startEnergy"] = self.scan_data[-1][0]
            self.scan_info["endEnergy"] = self.scan_data[-1][1]

            self.log.debug("Scan finished. data is: %s " % str(self.scan_data))
            self.emit("energyScanFinished", (self.scan_info,))
            self.emit("progressStop", ())

    def do_chooch(
        self, elt, edge, scan_directory, archive_directory, prefix, run_number=None
    ):
        symbol = "_".join((elt, edge))

        self.log.info("EnergyScan. executing do_chooch")

        scan_file_prefix = os.path.join(scan_directory, prefix)
        archive_file_prefix = os.path.join(archive_directory, prefix)

        if os.path.exists(scan_file_prefix + ".raw"):
            i = 1
            while os.path.exists(scan_file_prefix + "%d.raw" % i):
                i = i + 1
                scan_file_prefix += "_%d" % i
                archive_file_prefix += "_%d" % i

        scan_file_raw_filename = os.path.extsep.join((scan_file_prefix, "raw"))
        archive_file_raw_filename = os.path.extsep.join((archive_file_prefix, "raw"))

        scan_file_efs_filename = os.path.extsep.join((scan_file_prefix, "efs"))
        archive_file_efs_filename = os.path.extsep.join((archive_file_prefix, "efs"))

        scan_file_png_filename = os.path.extsep.join((scan_file_prefix, "png"))
        archive_file_png_filename = os.path.extsep.join((archive_file_prefix, "png"))

        try:
            if not os.path.exists(scan_directory):
                os.makedirs(scan_directory)
            if not os.path.exists(archive_directory):
                os.makedirs(archive_directory)
        except Exception:
            logging.getLogger("HWR").exception(
                "PX1EnergyScan: could not create results directory."
            )
            self.store_energy_scan()
            self.scanCommandFailed()
            return

        if not self.save_raw(scan_file_raw_filename, archive_file_raw_filename):
            logging.getLogger("HWR").exception(
                "PX1EnergyScan: could not save data raw file"
            )
            self.store_energy_scan()
            self.scanCommandFailed()
            return

        self.scan_info["scanFileFullPath"] = str(scan_file_raw_filename)

        # run chooch command
        self.log.info(
            "EnergyScan. running chooch %s %s %s %s"
            % (self.chooch_cmd, scan_file_efs_filename, elt, edge)
        )
        self.log.info(
            "   on success efs file should be saved : %s" % scan_file_efs_filename
        )
        try:
            p = subprocess.Popen(
                [
                    self.chooch_cmd,
                    scan_file_raw_filename,
                    elt,
                    edge,
                    scan_file_efs_filename,
                ],
                stdout=subprocess.PIPE,
            )
            chooch_result_lines = p.communicate()[0].split("\n")

            # there could messages in stdout. results are identified with -chooch_results- header line
            # produced by command (look for run_chooch in [..]/MXCuBE/tools
            next_lineno = 0
            found = 0
            for line in chooch_result_lines:
                next_lineno += 1
                if line == "chooch_results":
                    result_line = self.clean_chooch_line(
                        chooch_result_lines[next_lineno]
                    )
                    found = 1
                    break

            if found:
                self.log.debug("chooch got data back. %s" % result_line)
                result_data = eval(result_line)
                (
                    pk,
                    fppPeak,
                    fpPeak,
                    ip,
                    fppInfl,
                    fpInfl,
                    chooch_graph_data,
                ) = result_data
            else:
                self.store_energy_scan()
                logging.getLogger("HWR").error(
                    "Energy scan: Chooch cannot parse results"
                )
                return
            # scan_data = self.get_scan_data()
            # self.log.debug("running chooch with values: element=%s, element=%s" % (elt,edge))
            # self.log.debug("                   data is: %s" % str(scan_data))

            # result = PyChooch.calc(scan_data, elt, edge)
            # self.log.debug("          chooch returns : %s" % str(result))
            # pk, fppPeak, fpPeak, ip, fppInfl, fpInfl, chooch_graph_data = \
            #       result
        except Exception:
            import traceback

            self.log.debug(traceback.format_exc())
            self.store_energy_scan()
            logging.getLogger("HWR").error("Energy scan: Chooch failed")
            return

        self.log.info("EnergyScan. running chooch done")

        # scanData = self.get_scan_data()
        # logging.info('scanData %s' % scanData)
        # logging.info('PyChooch file %s' % PyChooch.__file__)
        # pk, fppPeak, fpPeak, ip, fppInfl, fpInfl, chooch_graph_data = PyChooch.calc(scanData, elt, edge, scanFile)

        rm = (pk + 30) / 1000.0
        pk = pk / 1000.0
        savpk = pk
        ip = ip / 1000.0

        self.thEdge = self.e_edge

        logging.getLogger("HWR").info(
            "th. Edge %s ; chooch results are pk=%f, ip=%f, rm=%f"
            % (self.thEdge, pk, ip, rm)
        )
        if math.fabs(self.thEdge - ip) > self.thEdgeThreshold:
            if (self.thEdge - ip) > 0.02:
                side = "below"
            else:
                side = "above"
            logging.info(
                "PX1EnergyScan. Theoretical edge too different from the one just determined. thEdgeThreshold = %.2f"
                % self.thEdgeThreshold
            )
            pk = 0
            ip = 0
            rm = self.thEdge + 0.03

            logging.getLogger("HWR").warning(
                "EnergyScan: calculated peak is %s theoretical value more than 20eV"
                % side
            )
            logging.getLogger("HWR").warning("   calculated = %s" % savpk)
            logging.getLogger("HWR").warning("  theoretical = %s" % self.thEdge)

        if not self.copy_efs(scan_file_efs_filename, archive_file_efs_filename):
            logging.getLogger("HWR").warning("  copy efs failed ")
            self.store_energy_scan()
            self.scanCommandFailed()
            return

        logging.getLogger("HWR").warning(
            "  efs file has been archived at %s" % archive_file_efs_filename
        )

        if pk > 0 and (not self.test_data_mode):
            self.move_egy_to_peak(pk)

        self.scan_info["peakEnergy"] = pk
        self.scan_info["inflectionEnergy"] = ip
        self.scan_info["remoteEnergy"] = rm
        self.scan_info["peakFPrime"] = fpPeak
        self.scan_info["peakFDoublePrime"] = fppPeak
        self.scan_info["inflectionFPrime"] = fpInfl
        self.scan_info["inflectionFDoublePrime"] = fppInfl
        self.scan_info["comments"] = ""

        self.scan_info["choochFileFullPath"] = scan_file_efs_filename
        self.scan_info["filename"] = archive_file_raw_filename
        # self.scan_info["workingDirectory"] = archive_directory

        logging.getLogger("HWR").warning(
            "  generating graph data from %s" % str(chooch_graph_data)
        )

        chooch_graph_x, chooch_graph_y1, chooch_graph_y2 = zip(*chooch_graph_data)
        chooch_graph_x = list(chooch_graph_x)
        for i in range(len(chooch_graph_x)):
            chooch_graph_x[i] = chooch_graph_x[i] / 1000.0

        logging.getLogger("HWR").info("PX1EnergScan. Saving png")

        # prepare to save png files
        title = "%10s  %6s  %6s\n%10s  %6.2f  %6.2f\n%10s  %6.2f  %6.2f" % (
            "energy",
            "f'",
            "f''",
            pk,
            fpPeak,
            fppPeak,
            ip,
            fpInfl,
            fppInfl,
        )

        scan_data = self.get_scan_data()

        fig = Figure(figsize=(15, 11))
        ax = fig.add_subplot(211)
        ax.set_title("%s\n%s" % (scan_file_efs_filename, title))
        ax.grid(True)
        ax.plot(*(zip(*scan_data)), **{"color": "black"})
        ax.set_xlabel("Energy")
        ax.set_ylabel("MCA counts")
        ax2 = fig.add_subplot(212)
        ax2.grid(True)
        ax2.set_xlabel("Energy")
        ax2.set_ylabel("")
        handles = []
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y1, color="blue"))
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y2, color="red"))
        canvas = FigureCanvasAgg(fig)

        escan_ispyb_path = HWR.beamline.session.path_to_ispyb(archive_file_png_filename)
        self.scan_info["jpegChoochFileFullPath"] = str(escan_ispyb_path)

        try:
            logging.getLogger("HWR").info(
                "Rendering energy scan and Chooch graphs to PNG file : %s",
                scan_file_png_filename,
            )
            canvas.print_figure(scan_file_png_filename, dpi=80)
        except Exception:
            logging.getLogger("HWR").exception("could not print figure")
        try:
            logging.getLogger("HWR").info(
                "Rendering energy scan and Chooch graphs to PNG file : %s",
                archive_file_png_filename,
            )
            canvas.print_figure(archive_file_png_filename, dpi=80)
        except Exception:
            logging.getLogger("HWR").exception("could not save figure")

        self.store_energy_scan()

        self.emit(
            "choochFinished",
            (
                pk,
                fppPeak,
                fpPeak,
                ip,
                fppInfl,
                fpInfl,
                rm,
                chooch_graph_x,
                chooch_graph_y1,
                chooch_graph_y2,
                title,
            ),
        )

        return (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            chooch_graph_x,
            chooch_graph_y1,
            chooch_graph_y2,
            title,
        )

    def clean_chooch_line(self, result_line):
        line = result_line.replace("nan", "0")
        return line

    def save_raw(self, scan_filename, archive_filename):
        try:
            scan_file_raw = open(scan_filename, "w")
            archive_file_raw = open(archive_filename, "w")
            self.log.info("EnergyScan. saving data in %s" % scan_filename)
            self.log.info("EnergyScan. archiving data in %s" % archive_filename)
        except Exception:
            logging.getLogger("HWR").exception(
                "EMBLEnergyScan: could not create results raw file"
            )
            return False

        scan_data = self.get_scan_data()
        for i in range(len(scan_data)):
            x = float(scan_data[i][0])
            y = float(scan_data[i][1])
            x = x < 1000 and x * 1000.0 or x  # convert to kEv if not done yet
            scan_file_raw.write("%f %f\r\n" % (x, y))
            archive_file_raw.write("%f %f\r\n" % (x, y))

        scan_file_raw.close()
        archive_file_raw.close()
        return True

    def copy_efs(self, from_file, to_file):
        try:
            with open(from_file) as ifd, open(to_file, "w") as ofd:
                ofd.write(ifd.read())
            return True
        except Exception:
            return False

    @task
    def store_energy_scan(self):
        if HWR.beamline.lims:
            scan_info = dict(self.scan_info)
            sample_id = scan_info["blSampleId"]
            scan_info.pop("blSampleId")

            self.log.debug("storing energy scan info in ISPyB")
            db_ret = HWR.beamline.lims.storeEnergyScan(scan_info)
            self.log.debug("stored %s" % str(db_ret))

            if sample_id is not None and db_ret is not None:
                scan_id = db_ret["energyScanId"]

                asoc = {"blSampleId": sample_id, "energyScanId": scan_id}
                HWR.beamline.lims.associate_bl_sample_and_energy_scan(asoc)

        if self.ruche_hwo:
            self.ruche_hwo.trigger_sync(self.escan_archivepng)

    # Elements commands
    def get_elements(self):
        elements = []
        try:
            for el in self["elements"]:
                elements.append({"symbol": el.symbol, "energy": el.energy})
        except IndexError:
            pass
        return elements

    # def getDefaultMadEnergies(self):
    #     # this does not seem to be used anywhere
    #     energies = []
    #     try:
    #         for el in self["mad"]:
    #             energies.append([float(el.energy), el.directory])
    #     except IndexError:
    #         pass
    #     return energies

    def get_scan_data(self):
        if self.test_data_mode:
            return self.simul_data
        return self.scan_data

    def load_data_file(self, filename):
        lines = open(filename).readlines()

        data = []
        in_header = True

        for line in lines:
            line = line.strip()
            try:
                x, y = map(float, line.split())
                # got good data... i
                in_header = False
            except Exception:
                if in_header:
                    continue
                else:
                    print("Wrong line in data file")

            data.append((x, y))

        self.log.debug(" loaded simulated data is \n%s" % str(data))
        return data


def test_hwo(scan):
    print("ELEMENTS:")
    print("---------")
    print(scan.get_elements())
    print(scan.get_scan_data())



import base64
import json

from PIL import Image

HTML_START = """<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html lang="en">
<head>
  <title>%s</title>
</head>
<body>\n"""

HTML_END = """</body>
</html>"""

COLOR_DICT = {
    "LightRed": "#FFCCCC",
    "Red": "#FE0000",
    "LightGreen": "#CCFFCC",
    "Green": "#007800",
}


def create_text(text, heading=None, color=None, bold=None):
    if heading:
        html_str = "<h%d>%s</h%d>\n" % (heading, text, heading)
    else:
        html_str = text
    return html_str


def create_image(image_path, width=None, height=None):
    html_str = '<img src="%s" title="%s"' % (image_path, image_path)
    if width:
        html_str += " width=%d" % width
    if height:
        html_str += " height=%d" % height

    html_str += "/>\n"
    return html_str


def create_html_start(title=""):
    return HTML_START % title


def create_html_end():
    return HTML_END


def create_table(table_header=None, table_cells=None, border_size=1):
    string_list = []
    string_list.append("<table border='%d'>" % border_size)
    if table_header:
        string_list.append("<tr>")
        header_str = ""
        for table_header in table_header:
            if table_header.startswith("bgcolor"):
                header_str += "<th %s</th>" % table_header
            else:
                header_str += "<th>%s</th>" % table_header
        header_str += "\n"
        string_list.append(header_str)

    if table_cells:
        for table_row in table_cells:
            if str(table_row[0]).startswith("bgcolor"):
                row_str = "<tr %s>" % str(table_row[0])
                table_row.pop(0)
            else:
                row_str = "<tr>"
            for cell in table_row:
                cell_str = str(cell)
                if cell_str.startswith("<td bgcolor"):
                    row_str += cell_str
                elif cell_str.startswith("bgcolor"):
                    row_str += "<td %s" % cell_str
                else:
                    row_str += "<td>%s</td>" % cell_str
            row_str += "</tr>\n"
            string_list.append(row_str)
    if table_header:
        string_list.append("</tr>")
    string_list.append("</table>")
    return string_list


def create_ref(ref_name, ref_text=None, hidden=True):
    if ref_text:
        return "<a href=#%s>%s</a>" % (ref_name, ref_text)
    else:
        return "<a href=#%s></a>" % ref_name


def create_toc(ref_list, title):
    toc_str = '<nav role="navigation" class="table-of-contents">'
    if title:
        toc_str += "<h2>%s</h2>" % title
    toc_str += "<ul>"
    for ref in ref_list:
        toc_str += '<li><a href="#%s">%s</a></li>' % (ref[0], ref[1])
    toc_str += "</ul></nav>"
    return toc_str


def create_json_images(image_list):
    json_item = {"type": "images", "items": []}
    for image in image_list:
        item = {}
        item["type"] = "image"
        item["suffix"] = image["filename"].split(".")[-1]
        item["title"] = image["title"]
        im = Image.open(image["filename"])
        item["xsize"] = im.size[0] / 2
        item["ysize"] = im.size[1] / 2
        item["value"] = base64.b64encode(open(image["filename"]).read())
        """
        if item.get("thumbnail_image_filename") is None:
            if thumbnailHeight is not None and thumbnailWidth is not None:
                item["thumbnailSuffix"] = pathToImage.split(".")[-1]
                item["thumbnailXsize"] = thumbnailHeight
                item["thumbnailYsize"] = thumbnailWidth
                item["thumbnailValue"] = base64.b64encode(open(pathToImage).read())
        else:
            item["thumbnailSuffix"] = pathToThumbnailImage.split(".")[-1]
            thumbnailIm = PIL.Image.open(pathToThumbnailImage)
            item["thumbnailXsize"] = thumbnailIm.size[0]
            item["thumbnailYsize"] = thumbnailIm.size[1]
            item["thumbnailValue"] = base64.b64encode(open(pathToThumbnailImage).read())
        """
        json_item["items"].append(item)
    return json_item


def generate_parallel_processing_report(mesh_scan_results, params_dict):
    json_dict = {"items": []}

    html_file = open(params_dict["html_file_path"], "w")
    html_file.write('<div align="CENTER">\n')

    if params_dict["lines_num"] > 1:
        json_dict["items"].append({"type": "title", "value": "Mesh scan results"})
        html_file.write(HTML_START % "Mesh scan results")
    else:
        html_file.write(HTML_START % "Line scan results")
        json_dict["items"].append({"type": "title", "value": "Line scan results"})

    html_file.write(create_image("parallel_processing_plot.png"))
    html_file.write("</br>")
    html_file.write(create_text("Scan parameters", heading=1))
    osc_range_per_line = params_dict["osc_range"] * (params_dict["images_per_line"] - 1)

    table_cells = [
        ("Number of lines", str(params_dict["lines_num"])),
        ("Frames per line", str(params_dict["images_per_line"])),
    ]
    if params_dict["lines_num"] > 1:
        table_cells.extend(
            (
                (
                    "Grid size",
                    "%d x %d microns"
                    % (
                        (params_dict["steps_x"] * params_dict["xOffset"] * 1000),
                        (params_dict["steps_y"] * params_dict["yOffset"] * 1000),
                    ),
                ),
                (
                    "Scan area",
                    "%d x %d microns"
                    % ((params_dict["dx_mm"] * 1000), (params_dict["dy_mm"] * 1000)),
                ),
                (
                    "Horizontal distance between frames",
                    "%d microns" % (params_dict["xOffset"] * 1000),
                ),
                (
                    "Vertical distance between frames",
                    "%d microns" % (params_dict["xOffset"] * 1000),
                ),
                ("Osciallation middle", "%.1f" % params_dict["osc_midle"]),
                ("Osciallation range per frame", "%.2f" % params_dict["osc_range"]),
                (
                    "Osciallation range per line",
                    "%.2f (from %.2f to %2.f)"
                    % (
                        osc_range_per_line,
                        (params_dict["osc_midle"] - osc_range_per_line / 2),
                        (params_dict["osc_midle"] + osc_range_per_line / 2),
                    ),
                ),
            )
        )
    table_rec = create_table(table_cells=table_cells, border_size=0)
    for row in table_rec:
        html_file.write(row)
    html_file.write("</br>")

    positions = mesh_scan_results.get("best_positions", [])
    if len(positions) > 0:
        html_file.write(create_text("Best position", heading=1))
        html_file.write("</br>")

        html_file.write('<font size="2">')
        table_cells = [
            [
                "%d" % positions[0]["index"],
                "<b>%.2f<b>" % positions[0]["score"],
                "<b>%d</b>" % positions[0]["spots_num"],
                "%.1f" % positions[0]["spots_resolution"],
                positions[0]["filename"],
                "%d" % (positions[0]["col"] + 0.5),
                "%d" % (positions[0]["row"] + 0.5),
            ]
        ]
        table_rec = create_table(
            [
                "Index",
                "<b>Score</b>",
                "<b>Number of spots</b>",
                "Resolution",
                "File name",
                "Column",
                "Row",
            ],
            table_cells,
        )
        for row in table_rec:
            html_file.write(row)
        html_file.write("</br>")

        if len(positions) > 1:
            html_file.write(create_text("All positions", heading=1))
            html_file.write("</br>")
            table_cells = []
            for position in positions[1:]:
                table_cells.append(
                    (
                        position["index"],
                        "<b>%.2f</b>" % position["score"],
                        "<b>%d</b>" % position["spots_num"],
                        "%.1f" % position["spots_resolution"],
                        position["filename"],
                        "%d" % (position["col"] + 0.5),
                        "%d" % (position["row"] + 0.5),
                    )
                )
            table_rec = create_table(
                [
                    "Index",
                    "<b>Score</b>",
                    "<b>Number of spots</b>",
                    "Resolution",
                    "File name",
                    "Column",
                    "Row",
                ],
                table_cells,
            )
            for row in table_rec:
                html_file.write(row)
            html_file.write("</br>")
        html_file.write("</font>")
    html_file.write("</div>\n")
    html_file.write(HTML_END)
    html_file.close()

    image = {"title": "plot", "filename": params_dict["cartography_path"]}
    json_dict["items"].append(create_json_images([image]))
    open(params_dict["json_file_path"], "w").write(json.dumps(json_dict, indent=4))



"""
FLEX HCD maintenance mockup.
"""

import ast

from mxcubecore.BaseHardwareObjects import HardwareObject

TOOL_FLANGE, TOOL_UNIPUCK, TOOL_SPINE, TOOL_PLATE, TOOL_LASER, TOOL_DOUBLE_GRIPPER = (
    0,
    1,
    2,
    3,
    4,
    5,
)

TOOL_TO_STR = {
    "Flange": TOOL_FLANGE,
    "Unipuck": TOOL_UNIPUCK,
    "Rotat": TOOL_SPINE,
    "Plate": TOOL_PLATE,
    "Laser": TOOL_LASER,
    "Double": TOOL_DOUBLE_GRIPPER,
}


class FlexHCDMaintenance(HardwareObject):
    __TYPE__ = "FLEX_HCD"
    NO_OF_LIDS = 3

    """
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self):
        self._sc = self.get_object_by_role("sample_changer")

    def get_current_tool(self):
        return self._sc.get_gripper()

    def _do_trash(self):
        """
        Trash sample

        :returns: Task id
        :rtype: int
        """
        return self._sc._do_trash()

    def _do_abort(self):
        """
        Abort current command

        :returns: None
        :rtype: None
        """
        return self._sc._do_abort()

    def _do_home(self):
        """
        Abort current command

        :returns: None
        :rtype: None
        """
        self._sc._do_abort()
        return self._sc.home()

    def _do_reset(self):
        """
        Reset sample changer

        :returns: None
        :rtype: None
        """
        self._sc._do_reset()

    def _do_defreeze_gripper(self):
        """
        :returns: None
        :rtype: None
        """
        self._sc.defreeze()

    def _do_change_gripper(self, args):
        """
        :returns: None
        :rtype: None
        """
        self._sc.change_gripper(gripper=args)
        self.emit("gripperChanged")

    def _do_reset_sample_number(self):
        """
        :returns: None
        :rtype: None
        """
        self._sc.reset_loaded_sample()

    def _update_global_state(self):
        state_dict, cmd_state, message = self.get_global_state()
        self.emit("globalStateChanged", (state_dict, cmd_state, message))

    def get_global_state(self):
        """ """
        state = self._sc._read_state()
        running = state in ("RUNNING",)

        state_dict = {"running": running, "state": state}

        cmd_state = {
            "home": True,
            "defreeze": True,
            "reset_sample_number": True,
            "change_gripper": True,
            "abort": True,
            "trash": True,
        }

        message = ""

        return state_dict, cmd_state, message

    def get_cmd_info(self):
        """return information about existing commands for this object
        the information is organized as a list
        with each element contains
        [ cmd_name,  display_name, category ]
        """
        """ [cmd_id, cmd_display_name, nb_args, cmd_category, description ] """

        cmd_list = [
            [
                "Actions",
                [
                    ["home", "Home", "Actions", None],
                    ["defreeze", "Defreeze gripper", "Actions", None],
                    ["reset_sample_number", "Reset sample number", "Actions", None],
                    ["abort", "Abort", "Actions", None],
                    ["trash", "Trash sample", "Actions", None],
                ],
            ],
        ]

        exclude_command_list = ast.literal_eval(
            self.get_property("exclude_commands", "[]")
        )
        new_command_list = []

        for command in cmd_list[0][1]:
            if command[0] not in exclude_command_list:
                new_command_list.append(command)

        cmd_list[0][1] = new_command_list

        try:
            grippers = self._sc.get_available_grippers()
        except Exception:
            pass
        else:
            gripper_cmd_list = []

            for gripper in grippers:
                arg = list(self._sc.gripper_types.keys())[
                    list(self._sc.gripper_types.values()).index(gripper)
                ]
                gripper_cmd_list.append(
                    [
                        "change_gripper",
                        gripper.title().replace("_", " "),
                        "Gripper",
                        arg,
                    ]
                )

            grippers_cmd = [
                "Gripper: %s" % self._sc.get_gripper().title().replace("_", " "),
                gripper_cmd_list,
            ]

            cmd_list.append(grippers_cmd)

        return cmd_list

    def send_command(self, cmd_name, args=None):
        if cmd_name in ["home"]:
            self._do_home()
        if cmd_name in ["defreeze"]:
            self._do_defreeze_gripper()
        if cmd_name in ["reset_sample_number"]:
            self._do_reset_sample_number()
        if cmd_name == "change_gripper":
            self._do_change_gripper(int(args))
        if cmd_name == "trash":
            self._do_trash()
        if cmd_name == "abort":
            self._do_abort()

        return True



import logging

from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__category__ = "General"


class EMBLPPUControl(HardwareObject):
    """
    Allows to restart processes on PPU
    """

    def __init__(self, name):
        super().__init__(name)

        self.all_status = None
        self.status_result = None
        self.restart_result = None

        self.error_state = None
        self.is_error = False
        self.file_transfer_in_error = False

        self.cmd_all_status = None
        self.cmd_furka_restart = None
        self.cmd_all_restart = None
        self.chan_all_status = None
        self.chan_file_info = None
        self.chan_all_restart = None

        self.msg = ""

        self.status_running = None
        self.restart_running = None
        self.execution_state = None

    def init(self):
        self.all_status = ""
        self.status_result = ""
        self.restart_result = ""
        self.execution_state = self.get_property("executionState")
        self.error_state = self.get_property("errorState")

        self.chan_all_status = self.get_channel_object("chanAllStatus")

        self.cmd_all_status = self.get_command_object("cmdAllStatus")
        self.cmd_all_restart = self.get_command_object("cmdAllRestart")
        self.cmd_furka_restart = self.get_command_object("cmdFurkaRestart")
        self.cmd_furka_restart("")

        self.get_status()

        self.chan_file_info = self.get_channel_object("chanFileInfo", optional=True)
        if self.chan_file_info is not None:
            self.chan_file_info.connect_signal("update", self.file_info_changed)

        self.connect(self.chan_all_status, "update", self.all_status_changed)

        self.chan_all_restart = self.get_channel_object("chanAllRestart")
        self.connect(self.chan_all_restart, "update", self.all_restart_changed)

    def all_status_changed(self, status):
        """
        Updates status
        :param status: str
        :return:
        """
        if self.status_running and not status:
            self.all_status = self.cmd_all_status.get()  # status
            self.update_status()
        self.status_running = status

    def all_restart_changed(self, status):
        """
        Updates status after all process restart
        :param status:
        :return:
        """
        if self.restart_running and not status:
            self.restart_result = self.cmd_all_restart.get()  # status
        self.restart_running = status

    def file_info_changed(self, values):
        """
        Updated information about transferred frames
        values is a list of 3 values, where the last one indicates the number
        of dropped frames
        :param values:
        :return:
        """
        values = list(values)
        # if len(values) == 2:
        #    value = values[1]
        # else:
        #    value = values[0]

        self.file_transfer_in_error = values[2] > 0
        self.emit("fileTranferStatusChanged", (values))

        self.is_error = (
            self.all_status.startswith(self.error_state) or self.file_transfer_in_error
        )

        if self.file_transfer_in_error:
            self.emit("ppuStatusChanged", self.is_error, "File tansfer in error")

    def get_status(self):
        """
        Returns status
        :return: boolean, str
        """
        self.cmd_all_status("")
        return self.is_error, self.all_status

    def update_status(self):
        """
        Update status method
        :return:
        """
        self.is_error = (
            self.all_status.startswith(self.error_state) or self.file_transfer_in_error
        )

        if self.all_status.startswith(self.error_state):
            msg_list = self.all_status.split("\n")
            logging.getLogger("GUI").error("PPU control is in Error state!")
            if len(msg_list) > 1:
                for msg_line in msg_list:
                    if msg_line:
                        msg = "PPU control: %s" % msg_line
                        logging.getLogger("GUI").error(msg)
        else:
            msg = "PPUControl: %s" % self.all_status
            logging.getLogger("HWR").debug(msg)

        self.msg = (
            "Restart result:\n\n%s\n\n" % self.restart_result
            + "All status result:\n\n%s\n" % self.all_status
        )

        self.emit("ppuStatusChanged", self.is_error, self.msg)

        return self.is_error, self.all_status

    def restart_all(self):
        """
        Restarts all processes
        :return:
        """
        self.emit("ppuStatusChanged", False, "Restarting.... ")
        self.cmd_all_restart("")
        self.get_status()

    def re_emit_values(self):
        """
        Reemits signals
        :return:
        """
        self.emit("ppuStatusChanged", self.is_error, self.msg)



#
#  Project name: MXCuBEv
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""TINEMotor class defines motor in the TINE control system"""

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "Motor"


TINE_STATE_DICT = {
    0: HardwareObjectState.READY,
    "ready": HardwareObjectState.READY,
    "busy": HardwareObjectState.BUSY,
    "error": HardwareObjectState.FAULT,
}


class TINEMotor(AbstractMotor):
    """TINEMotor class defines motor in the TINE control system"""

    def __init__(self, name):
        AbstractMotor.__init__(self, name)

        self.chan_position = None
        self.chan_state = None
        self.chan_limits = None
        self.cmd_set_position = None
        self.cmd_stop_axis = None
        self.cmd_set_online = None
        self.step_limits = None

    def init(self):
        """Connects to all Tine channels and commands"""

        self.chan_limits = self.get_channel_object("axisLimits", optional=True)
        if self.chan_limits is not None:
            self.chan_limits.connect_signal("update", self.update_limits)
            self.update_limits(self.chan_limits.get_value())
        else:
            try:
                if self.get_property("default_limits"):
                    self.update_limits(eval(self.get_property("default_limits")))
            except Exception:
                pass

        self.chan_position = self.get_channel_object("axisPosition")
        if self.chan_position is not None:
            self.chan_position.connect_signal("update", self.update_tine_value)
        self.update_value(self.chan_position.get_value())

        self.chan_state = self.get_channel_object("axisState", optional=True)
        if self.chan_state is not None:
            self.chan_state.connect_signal("update", self.update_tine_state)

        self.cmd_set_position = self.get_command_object("setPosition")
        if self.cmd_set_position:
            self.cmd_set_position.connect("connected", self.connected)
            self.cmd_set_position.connect("disconnected", self.disconnected)

        self.cmd_stop_axis = self.get_command_object("stopAxis")
        if self.cmd_stop_axis:
            self.cmd_stop_axis.connect("connected", self.connected)
            self.cmd_stop_axis.connect("disconnected", self.disconnected)

        self.cmd_set_online = self.get_command_object("setOnline")

        # NBNB TODO change config from 'epsilon' to 'tolerance'?
        self._tolerance = self.get_property("epsilon")

        try:
            self.step_limits = eval(self.get_property("stepLimits"))
        except Exception:
            pass

    def connected(self):
        """
        Sets ready
        :return:
        """
        self.update_state(HardwareObjectState.READY)

    def disconnected(self):
        """
        Sets not ready
        :return:
        """
        self.update_state(HardwareObjectState.OFF)

    def connect_notify(self, signal):
        """
        :param signal: signal
        :type signal: signal
        """
        if self.connected():
            if signal == "stateChanged":
                self.update_state()
            elif signal == "limitsChanged":
                self.update_limits()
            elif signal == "valueChanged":
                self.update_value()

    def get_step_limits(self):
        """Returns step limits"""
        return self.step_limits

    # def get_position(self):
    #    return self.chan_position.get_value()

    def get_value(self):
        return self.chan_position.get_value()

    def get_state(self):
        """Get HardwareObject state"""
        # NNBNB TODO map channel states to all HardwareObject states
        # TODO add treatment of specific_states
        if self.chan_state is None:
            return HardwareObjectState.READY
        else:
            return self.convert_state(self.chan_state.get_value())

    def convert_state(self, tine_state):
        if type(tine_state) in (tuple, list):
            tine_state = tine_state[0]
        return TINE_STATE_DICT.get(tine_state, HardwareObjectState.UNKNOWN)

    def abort(self):
        """Stops motor movement"""
        self.cmd_stop_axis()

    def _set_value(self, value):
        """
        Main move method
        :param value: float
        :return:
        """
        if self.chan_state is not None:
            self.update_state(self.STATES.BUSY)
            self.chan_state.set_old_value("moving")
        if self.cmd_set_position:
            self.cmd_set_position(value)
        else:
            self.chan_position.set_value(value)

    def update_tine_value(self, value=None):
        """Updates motor position"""
        if type(value) in (list, tuple):
            value = value[0]
        super(TINEMotor, self).update_value(value)

    def update_tine_state(self, value):
        value = self.convert_state(value)
        super(TINEMotor, self).update_state(value)

    def get_motor_mnemonic(self):
        """
        Returns motor mnemonic
        :return:
        """
        return "TINEMotor"

    def enable_motor(self):
        """
        Enables motor
        """
        if self.cmd_set_online:
            self.cmd_set_online(1)
            gevent.sleep(2)

    def disable_motor(self):
        """
        Disables motor
        """
        if self.cmd_set_online:
            self.cmd_set_online(0)
            gevent.sleep(2)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import copy
import logging
import time

try:
    import epics
except ImportError:
    logging.getLogger("HWR").warning("EPICS support not available.")

from mxcubecore import Poller
from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
)
from mxcubecore.dispatcher import saferef

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


class EpicsCommand(CommandObject):
    """Epics Command"""

    def __init__(self, name, pv_name, username=None, args=None, **kwargs):
        CommandObject.__init__(self, name, username, **kwargs)

        self.pv_name = pv_name
        self.read_as_str = kwargs.get("read_as_str", False)
        self.auto_monitor = kwargs.get("auto_monitor", True)
        self.pollers = {}
        self.__value_changed_callback_ref = None
        self.__timeout_callback_ref = None

        if args is None:
            self.arg_list = ()
        else:
            # not very nice...
            args = str(args)
            if not args.endswith(","):
                args += ","
            self.arg_list = eval("(" + args + ")")

        if len(self.arg_list) > 1:
            logging.getLogger("HWR").error(
                "EpicsCommand: ftm only scalar arguments are supported."
            )
            return

        logging.getLogger("HWR").debug(
            "EpicsCommand: creating pv %s: read_as_str = %s",
            self.pv_name,
            self.read_as_str,
        )
        self.pv = epics.PV(pv_name, auto_monitor=self.auto_monitor)
        time.sleep(0.01)
        self.pv_connected = self.pv.connect(timeout=0.1)

        if self.pv_connected:
            self.value_changed(self.pv.get(as_string=self.read_as_str, timeout=0.1))
        else:
            logging.getLogger("HWR").error(
                "EpicsCommand: Error connecting to pv %s.", self.pv_name
            )

    def __call__(self, *args, **kwargs):
        self.emit("commandBeginWaitReply", (str(self.name()),))

        if len(args) > 0 and len(self.arg_list) > 0:
            # arguments given both given in command call _AND_ in the xml file
            logging.getLogger("HWR").error(
                "%s: cannot execute command with arguments when 'args' is defined from XML",
                str(self.name()),
            )
            self.emit("commandFailed", (-1, str(self.name())))
            return
        elif len(args) == 0 and len(self.arg_list) > 0:
            # no argument given in the command call but inside the xml file -> use the
            # default argument from the xml file
            args = self.arg_list

        if self.pv is not None:
            if len(args) == 0:
                # no arguments available -> get the pv's current value
                try:
                    ret = self.pv.get(as_string=self.read_as_str, timeout=0.2)
                    if ret is None:
                        ret = self.reconnect()
                except TypeError:
                    # When a cached info is lost internally Epics return a TypeError
                    ret = self.reconnect()
                    if ret is not None:
                        self.emit("commandReplyArrived", (ret, str(self.name())))
                        return ret
                except Exception as e:
                    logging.getLogger("HWR").error(
                        "%s: an error occured when getting value with Epics command %s",
                        str(self.name()),
                        self.pv_name,
                    )
                else:
                    self.emit("commandReplyArrived", (ret, str(self.name())))
                    return ret
            else:
                # use the given argument to change the pv's value
                try:
                    value = args[0]
                    wait = kwargs.get("wait", False)
                    self.pv.put(value, wait=wait)
                except:
                    logging.getLogger("HWR").error(
                        "%s: an error occured when putting a value with Epics command %s",
                        str(self.name()),
                        self.pv_name,
                    )
                else:
                    self.emit("commandReplyArrived", (0, str(self.name())))
                    return 0
        self.emit("commandFailed", (-1, str(self.name())))

    def value_changed(self, value):
        try:
            callback = self.__value_changed_callback_ref()
        except Exception:
            pass
        else:
            if callback is not None:
                callback(value)

    def on_polling_error(self, exception, poller_id):
        # try to reconnect the pv
        self.pv.connect()

        poller = Poller.get_poller(poller_id)
        if poller is not None:
            try:
                poller.restart(1000)
            except Exception:
                pass

    def get_pv_value(self):
        """wrapper function to pv.get() in order to supply additional named parameter"""
        return self.pv.get(as_string=self.read_as_str)

    def poll(
        self,
        polling_time=500,
        arguments_list=(),
        value_changed_callback=None,
        timeout_callback=None,
        direct=True,
        compare=True,
    ):
        self.__value_changed_callback_ref = saferef.safe_ref(value_changed_callback)

        # store the call to get as a function object
        poll_cmd = self.get_pv_value

        Poller.poll(
            poll_cmd,
            copy.deepcopy(arguments_list),
            polling_time,
            self.value_changed,
            self.on_polling_error,
            compare,
        )

    def stop_polling(self):
        pass

    def abort(self):
        pass

    def is_connected(self):
        return self.pv_connected

    def reconnect(self):
        epics.ca._cache.clear()
        # Reconnect PV
        self.pv = epics.PV(self.pv_name, auto_monitor=self.auto_monitor)
        self.pv_connected = self.pv.connect(timeout=0.2)
        # Return the result of get()
        ret = self.pv.get(as_string=self.read_as_str, timeout=0.2)
        return ret


class EpicsChannel(ChannelObject):
    """Emulates an *Epics channel* with an EpicsCommand +polling"""

    def __init__(self, name, command, username=None, polling=None, args=None, **kwargs):
        ChannelObject.__init__(self, name, username, **kwargs)
        self.command = EpicsCommand(
            name + "_internalCmd", command, username, args, **kwargs
        )
        try:
            self.polling = int(polling)
        except Exception:
            self.polling = None
        else:
            self.command.poll(self.polling, self.command.arg_list, self.value_changed)

    def value_changed(self, value):
        self.emit("update", value)

    def get_value(self):
        return self.command()

    def set_value(self, value):
        self.command(value)

    def is_connected(self):
        return self.command.is_connected()



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


from mxcubecore.HardwareObjects.abstract.AbstractTransmission import (
    AbstractTransmission,
)


class Attenuators(AbstractTransmission):
    def __init__(self, name):
        AbstractTransmission.__init__(self, name)

        self.chan_att_value = None
        self.chan_att_state = None
        self.chan_att_limits = None

    def init(self):
        self.chan_att_value = self.get_channel_object("chanAttValue")
        self.chan_att_value.connect_signal("update", self.value_changed)
        self.value_changed(self.chan_att_value.get_value())
        self.chan_att_state = self.get_channel_object("chanAttState")
        self.chan_att_state.connect_signal("update", self.state_changed)
        self.chan_att_limits = self.get_channel_object("chanAttLimits")
        self.chan_att_limits.connect_signal("update", self.limits_changed)

        self.re_emit_values()

    def state_changed(self, state):
        self._state = state
        self.emit("stateChanged", self._state)

    def value_changed(self, value):
        self._value = value
        self.emit("valueChanged", self._value)

    def limits_changed(self, value):
        self._limits = value
        self.emit("limitsChanged", (self._limits,))

    def _set_value(self, value):
        self.chan_att_value.set_value(value)



import logging
import os
import shutil
import time

import gevent
import gevent.event

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class XRFSpectrum(HardwareObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_data = None
        self.calib_data = None
        self.energy_spectrum_args = None

    def init(self):
        self.scanning = None
        self.ready_event = gevent.event.Event()

        try:
            self.doSpectrum.connect_signal(
                "commandBeginWaitReply", self.spectrumCommandStarted
            )
            self.doSpectrum.connect_signal("commandFailed", self.spectrumCommandFailed)
            self.doSpectrum.connect_signal(
                "commandAborted", self.spectrumCommandAborted
            )
            self.doSpectrum.connect_signal("commandReady", self.spectrumCommandReady)
            self.doSpectrum.connect_signal(
                "commandNotReady", self.spectrumCommandNotReady
            )
        except AttributeError as diag:
            logging.getLogger().warning(
                "XRFSpectrum: error initializing XRF spectrum (%s), probably not using SPEC macros"
                % str(diag)
            )
            self.doSpectrum = None
        else:
            self.doSpectrum.connect_signal("connected", self.sConnected)
            self.doSpectrum.connect_signal("disconnected", self.sDisconnected)

        if HWR.beamline.lims is None:
            logging.getLogger().warning(
                "XRFSpectrum: you should specify the database hardware object"
            )
        self.spectrumInfo = None

        self.ctrl_hwobj = self.get_object_by_role("controller")
        self.mca_hwobj = self.ctrl_hwobj.mca
        # self.mca_hwobj = self.get_object_by_role("mca")
        # if self.mca_hwobj:
        #    self.mca_hwobj.set_calibration(calib_cf=self.mca_hwobj.calib_cf)

        self.archive_path = self.get_property("archive_path")
        if not self.archive_path:
            self.archive_path = "/data/pyarch/"

        self.cfg_path = self.get_property("cfg_path")
        if not self.cfg_path:
            self.cfg_path = "/users/blissadm/local/beamline_configuration/misc"

        if self.is_connected():
            self.sConnected()

    def is_connected(self):
        try:
            return self.doSpectrum.is_connected()
        except Exception:
            return False

    # Handler for spec connection
    def sConnected(self):
        self.emit("connected", ())
        # curr = self.getSpectrumParams()

    # Handler for spec disconnection
    def sDisconnected(self):
        self.emit("disconnected", ())

    # Energy spectrum commands
    def canSpectrum(self):
        if not self.is_connected():
            return False
        return self.doSpectrum is not None

    def startXrfSpectrum(
        self,
        ct,
        directory,
        archive_directory,
        prefix,
        session_id=None,
        blsample_id=None,
    ):
        self.spectrumInfo = {"sessionId": session_id, "blSampleId": blsample_id}
        self.spectrumCommandStarted()
        if not os.path.isdir(directory):
            logging.getLogger("user_level_log").debug(
                "XRFSpectrum: creating directory %s", directory
            )
            try:
                os.makedirs(directory)
            except OSError as diag:
                logging.getLogger("user_level_log").error(
                    "XRFSpectrum: error creating directory %s (%s)"
                    % (directory, str(diag))
                )
                self.spectrumStatusChanged("Error creating directory")
                return False

        curr = self.getSpectrumParams()

        try:
            curr["escan_dir"] = directory
            curr["escan_prefix"] = prefix
        except TypeError:
            curr = {}
            curr["escan_dir"] = directory
            curr["escan_prefix"] = prefix

        if not archive_directory:
            a = directory.split(os.path.sep)
            suffix_path = os.path.join(*a[4:])
            if "inhouse" in a:
                archive_directory = os.path.join(self.archive_path, a[2], suffix_path)
            else:
                archive_directory = os.path.join(self.archive_path, a[4], a[3], *a[5:])

        if not os.path.exists(archive_directory):
            try:
                logging.getLogger("user_level_log").debug(
                    "XRFSpectrum: creating %s", archive_directory
                )
                os.makedirs(archive_directory)
            except OSError as diag:
                logging.getLogger().error(
                    "XRFSpectrum: error creating directory %s (%s)",
                    (archive_directory, str(diag)),
                )
                self.spectrumStatusChanged("Error creating directory")
                return False

        _pattern = "%s_%s_%%02d" % (prefix, time.strftime("%d_%b_%Y"))
        filename_pattern = os.path.join(directory, _pattern)

        filename_pattern = os.path.extsep.join((filename_pattern, "dat"))
        filename = filename_pattern % 1
        fileprefix = _pattern % 1

        i = 2
        while os.path.isfile(filename):
            filename = filename_pattern % i
            fileprefix = _pattern % i
            i = i + 1

        archive_path = os.path.join(archive_directory, fileprefix)
        self.spectrumInfo["filename"] = filename
        self.spectrumInfo["scanFileFullPath"] = os.path.extsep.join(
            (archive_path, "dat")
        )
        self.spectrumInfo["jpegScanFileFullPath"] = os.path.extsep.join(
            (archive_path, "png")
        )
        self.spectrumInfo["annotatedPymcaXfeSpectrum"] = os.path.extsep.join(
            (archive_path, "html")
        )
        self.spectrumInfo["fittedDataFileFullPath"] = archive_path + "_peaks.csv"
        self.spectrumInfo["exposureTime"] = ct

        logging.getLogger("user_level_log").debug(
            "XRFSpectrum: archive file is %s", self.spectrumInfo["jpegScanFileFullPath"]
        )
        gevent.spawn(self.reallyStartXrfSpectrum, ct, filename)

        return True

    def reallyStartXrfSpectrum(self, ct, filename):
        try:
            res = self._doSpectrum(ct, filename, wait=True)
        except RuntimeError as err:
            msg = f"XRFSpectrum: problem calling procedure, {err}"
            logging.getLogger("user_level_log").exception(msg)
            self.spectrumStatusChanged("Error problem with spectrum procedure")
        else:
            self.spectrumCommandFinished(res)

    def cancelXrfSpectrum(self, *args):
        if self.scanning:
            self.doSpectrum.abort()

    def spectrumCommandReady(self):
        if not self.scanning:
            self.emit("xrfSpectrumReady", (True,))
            # self.emit('xrfScanReady', (True,))

    def spectrumCommandNotReady(self):
        if not self.scanning:
            self.emit("xrfSpectrumReady", (False,))

    def spectrumCommandStarted(self, *args):
        self.spectrumInfo["startTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.scanning = True
        self.emit("xrfSpectrumStarted", ())

    def spectrumCommandFailed(self, *args):
        self.spectrumInfo["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.scanning = False
        self.storeXrfSpectrum()
        self.emit("xrfSpectrumFailed", ())
        self.ready_event.set()

    def spectrumCommandAborted(self, *args):
        self.scanning = False
        self.emit("xrfSpectrumFailed", ())
        self.ready_event.set()

    def spectrumCommandFinished(self, result):
        self.spectrumInfo["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        logging.getLogger().debug("XRFSpectrum: XRF spectrum result is %s" % result)
        self.scanning = False
        if result is not False:
            fname = self.spectrumInfo["filename"].replace(".dat", ".raw")
            self.mca_hwobj.datafile = str(fname)
            mcaData = self.mca_hwobj.read_roi_data(save_data=True)
            mcaCalib = self.mca_hwobj.calibration
            mcaConfig = {}
            self.spectrumInfo["beamTransmission"] = (
                HWR.beamline.transmission.get_value()
            )
            self.spectrumInfo["energy"] = HWR.beamline.energy.get_value()
            if HWR.beamline.flux:
                self.spectrumInfo["flux"] = HWR.beamline.flux.get_value()
            self.beamsize = self.get_object_by_role("beamsize")
            if self.beamsize:
                bsX = self.beamsize.get_size(self.beamsize.get_value().name)
                self.spectrumInfo["beamSizeHorizontal"] = bsX
                self.spectrumInfo["beamSizeVertical"] = bsX
            mcaConfig["att"] = self.spectrumInfo["beamTransmission"]
            mcaConfig["energy"] = self.spectrumInfo["energy"]
            mcaConfig["bsX"] = self.spectrumInfo["beamSizeHorizontal"]
            mcaConfig["bsY"] = self.spectrumInfo["beamSizeVertical"]
            roi = self.ctrl_hwobj.mca.get_roi()
            # roi = self.mca_hwobj.get_roi()
            mcaConfig["min"] = roi["chmin"]
            mcaConfig["max"] = roi["chmax"]
            mcaConfig["legend"] = self.spectrumInfo["annotatedPymcaXfeSpectrum"]
            mcaConfig["htmldir"], _ = os.path.split(mcaConfig["legend"])
            mcaConfig["file"] = self._get_cfgfile(self.spectrumInfo["energy"])
            try:
                self.set_data(mcaData, mcaCalib, mcaConfig)
            except Exception:
                self.emit("xrfSpectrumFinished", (mcaData, mcaCalib, mcaConfig))

            # here move the png file
            pf = self.spectrumInfo["filename"].split(".")
            pngfile = os.path.extsep.join((pf[0], "png"))
            if os.path.isfile(pngfile) is True:
                try:
                    shutil.copyfile(pngfile, self.spectrumInfo["jpegScanFileFullPath"])
                except Exception:
                    logging.getLogger().error("XRFSpectrum: cannot copy %s", pngfile)

            # copy raw data file to the archive directory
            try:
                shutil.copyfile(fname, self.spectrumInfo["scanFileFullPath"])
            except Exception:
                logging.getLogger().error(
                    "XRFSpectrum: cannot copy %s", self.spectrumInfo["filename"]
                )

            logging.getLogger().debug("finished %r", self.spectrumInfo)
            self.storeXrfSpectrum()

            # copy csv file in the raw data directory
            try:
                ff = self.spectrumInfo["filename"].replace(".dat", "_peaks.csv")
                shutil.copyfile(self.spectrumInfo["fittedDataFileFullPath"], ff)
            except Exception:
                logging.getLogger().error("XRFSpectrum: cannot copy %s", ff)
        else:
            self.spectrumCommandFailed()
        self.ready_event.set()

    def spectrumStatusChanged(self, status):
        self.emit("xrfScanStatusChanged", (status,))
        self.emit("xrfSpectrumStatusChanged", (status,))

    def storeXrfSpectrum(self):
        logging.getLogger().debug("db connection %r", HWR.beamline.lims)
        logging.getLogger().debug("spectrum info %r", self.spectrumInfo)
        if HWR.beamline.lims is None:
            return
        try:
            session_id = int(self.spectrumInfo["sessionId"])
        except Exception:
            return
        blsampleid = self.spectrumInfo["blSampleId"]

        db_status = HWR.beamline.lims.store_xfe_spectrum(self.spectrumInfo)

    def updateXrfSpectrum(self, spectrum_id, jpeg_spectrum_filename):
        pass

    def getSpectrumParams(self):
        if self.energy_spectrum_args:
            try:
                self.curr = self.energy_spectrum_args.get_value()
                return self.curr
            except Exception:
                logging.getLogger().exception(
                    "XRFSpectrum: error getting xrfspectrum parameters"
                )
                self.spectrumStatusChanged("Error getting xrfspectrum parameters")
                return False
        else:
            return True

    def setSpectrumParams(self, pars):
        self.energy_spectrum_args.set_value(pars)

    def _get_cfgfile(self, energy):
        if energy > 12.0:
            cfgname = "15"
        elif energy > 10.0:
            cfgname = "12"
        elif energy > 7.0:
            cfgname = "10"
        else:
            cfgname = "7"
        return os.path.join(self.cfg_path, "%skeV.cfg" % cfgname)

    def _doSpectrum(self, ct, filename, wait=True):
        if not ct:
            ct = 5
        safshut = self.get_object_by_role("safety_shutter")

        # stop the procedure if hutch not searched
        stat = safshut.getShutterState()
        if stat == "disabled":
            logging.getLogger("user_level_log").exception(
                "XRFSpectrum: hutch not searched, exiting"
            )
            return False

        fluodet_ctrl = self.get_object_by_role("fluodet_ctrl")
        fluodet_ctrl.actuatorIn()
        # put the beamstop in
        try:
            self.ctrl_hwobj.diffractometer.set_phase("DataCollection", wait=True)
        except Exception:
            pass

        # open the safety and the fast shutter
        safshut.openShutter()
        init_transm = HWR.beamline.transmission.get_value()
        logging.getLogger("user_level_log").info(
            "Looking for maximum attenuation, please wait"
        )
        ret = self._findAttenuation(ct)
        self.ctrl_hwobj.diffractometer.msclose()
        fluodet_ctrl.actuatorOut()
        HWR.beamline.transmission.set_value(init_transm)
        return ret

    def _findAttenuation(self, ct):
        table = self.get_property("transmission_table")
        if table:
            tf = []
            for i in table.split(","):
                tf.append(float(i))
        else:
            tf = [0.1, 0.2, 0.3, 0.9, 1.3, 1.9, 2.6, 4.3, 6, 8, 12, 24, 36, 50]

        min_cnt = self.get_property("min_cnt")
        max_cnt = self.get_property("max_cnt")
        self.mca_hwobj.set_roi(2, 15, channel=1)
        fname = self.spectrumInfo["filename"].replace(".dat", ".raw")
        self.mca_hwobj.set_presets(erange=1, ctime=ct, fname=fname)

        # put in max attenuation
        HWR.beamline.transmission.set_value(0)

        self.ctrl_hwobj.diffractometer.msopen()
        self.mca_hwobj.start_acq()
        time.sleep(ct)
        ic = sum(self.mca_hwobj.read_roi_data()) / ct
        print(ic)
        if ic > max_cnt:
            self.ctrl_hwobj.diffractometer.msclose()
            logging.getLogger("user_level_log").exception(
                "The detector is saturated, giving up."
            )
            return False

        for i in tf:
            self.mca_hwobj.clear_spectrum()
            logging.getLogger("user_level_log").info("Setting transmission to %g" % i)
            HWR.beamline.transmission.set_value(i)
            self.mca_hwobj.start_acq()
            time.sleep(ct)
            ic = sum(self.mca_hwobj.read_roi_data()) / ct
            print(ic)
            if ic > min_cnt:
                self.ctrl_hwobj.diffractometer.msclose()
                self.spectrumInfo["beamTransmission"] = (
                    HWR.beamline.transmission.get_value()
                )
                logging.getLogger("user_level_log").info(
                    "Transmission used for spectra: %g"
                    % self.spectrumInfo["beamTransmission"]
                )
                break

        self.spectrumInfo["beamTransmission"] = HWR.beamline.transmission.get_value()
        self.ctrl_hwobj.diffractometer.msclose()
        if ic < min_cnt:
            logging.getLogger("user_level_log").exception(
                "Could not find satisfactory attenuation (is the mca properly set up?), giving up."
            )
            return False

        return True



import logging
import os
import time

from mxcubecore.HardwareObjects import Session
from mxcubecore.model import queue_model_objects


class ALBASession(Session.Session):
    def get_base_data_directory(self):
        """
        Returns the base data directory for ALBA
        In ALBA the base directory already includes the user
        home directory. So
        information into account, such as if the current user
        is inhouse.

        :returns: The base data path.
        :rtype: str
        """
        if self.session_start_date:
            start_time = self.session_start_date.split(" ")[0].replace("-", "")
        else:
            start_time = time.strftime("%Y%m%d")

        # directory = os.path.join(self.base_directory, self.get_proposal(), 'DATA', start_time)
        if self.base_directory is not None:
            directory = os.path.join(self.base_directory, start_time)
        else:
            directory = "/tmp"
        return directory

    def get_archive_directory(self, directory=None):
        if directory is None:
            thedir = self.get_base_data_directory()
        else:
            thedir = directory

        parts = thedir.split(os.path.sep)
        user_dir = parts[5]
        session_date = parts[6]
        # remove RAW_DATA from da
        try:
            more = parts[8:]
        except Exception:
            more = []

        archive_dir = os.path.join(
            self.base_archive_directory, user_dir, session_date, *more
        )
        # if 'RAW_DATA' in thedir:
        #    thedir = thedir.replace('RAW_DATA','ARCHIVE')
        # else:
        #    thedir = os.path.join(thedir, 'ARCHIVE')

        logging.getLogger("HWR").debug(
            "ALBASession. returning archive directory: %s" % archive_dir
        )
        return archive_dir

    def set_ldap_homedir(self, homedir):
        self.base_directory = homedir
        self.base_process_directory = homedir

        queue_model_objects.PathTemplate.set_data_base_path(self.base_directory)


def test_hwo(hwo):
    print(hwo.get_base_data_directory())
    print(hwo.get_process_directory())
    print(hwo.get_archive_directory())



"""
QueueManager, handles the execution of the MxCuBE queue. It is implemented
as a hardware object and is configured by an XML file. See the example of the
XML configuration for more details.

The Queue manager acts as both the controller of execution and as the root/
container of the queue, note the inheritance from QueueEntryContainer. See the
documentation for the queue_entry module for more information.
"""

import logging
import traceback

import gevent

from mxcubecore import queue_entry
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model.queue_model_enumerables import CENTRING_METHOD
from mxcubecore.queue_entry import base_queue_entry
from mxcubecore.queue_entry.base_queue_entry import QUEUE_ENTRY_STATUS

QueueEntryContainer = base_queue_entry.QueueEntryContainer


class QueueManager(HardwareObject, QueueEntryContainer):
    def __init__(self, name):
        HardwareObject.__init__(self, name)
        QueueEntryContainer.__init__(self)
        self.centring_method = CENTRING_METHOD.NONE
        self._root_task = None
        self._paused_event = gevent.event.Event()
        self._paused_event.set()
        self._current_queue_entry = None
        self._current_queue_entries = []
        self._running = False
        self._disable_collect = False
        self._is_stopped = False

    def init(self):
        site_entry_path = self.get_property("site_entry_path")
        if site_entry_path:
            queue_entry.import_queue_entries(site_entry_path.split(","))
        else:
            queue_entry.import_queue_entries()

    def __getstate__(self):
        d = dict(self.__dict__)
        d["_root_task"] = None
        d["_paused_event"] = None
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self._paused_event = gevent.event.Event()

    @property
    def current_queue_entries(self):
        return self._current_queue_entries

    def enqueue(self, queue_entry):
        """
        Method inherited from QueueEntryContainer, enqueues the QueueEntry
        object <queue_entry>.

        :param queue_entry: QueueEntry to add
        :type queue_entry: QueueEntry

        :returns: None
        :rtype: NoneType
        """

        queue_entry.set_queue_controller(self)
        super(QueueManager, self).enqueue(queue_entry)

    def execute(self, entry=None):
        """
        Starts execution of the queue.

        Runs the entire queue or a single entry (entry is set) as one "run" of
        the queue and  manages the various states such as "running", "paused"
        and "stopped".

        :param entry: Optional, queue_entry to run
        :type entry: QueueEntry
        :raises: RuntimeError, if the queue is already running when called
        """
        if self._running:
            raise RuntimeError("Can't call execute on a queue that is already running")

        if not self.is_disabled():
            # If no entry is passed run all entries in the queue
            # otherwise, run only the entry given
            self.emit("statusMessage", ("status", "Queue running", "running"))
            self._is_stopped = False
            self._running = True

            if not entry:
                self._current_queue_entries = []
                self._set_in_queue_flag()
                self._root_task = gevent.spawn(self.__execute_task)
            else:
                task = gevent.spawn(self.__execute_entry, entry)
                task.link((lambda _t: self._queue_end()))

    def _set_in_queue_flag(self):
        """
        Methods iterates over all queue entries and sets in_queue flag for
        DataCollectionQueue entries
        """
        self.entry_list = []

        def get_data_collection_list(entry):
            for child in entry._queue_entry_list:
                if (
                    (
                        isinstance(child, queue_entry.DataCollectionQueueEntry)
                        or isinstance(
                            child, queue_entry.CharacterisationGroupQueueEntry
                        )
                    )
                    and not child.get_data_model().is_executed()
                    and child.is_enabled()
                ):
                    self.entry_list.append(child)
                get_data_collection_list(child)

        for qe in self._queue_entry_list:
            get_data_collection_list(qe)

        if len(self.entry_list) > 1:
            for index, entry in enumerate(self.entry_list[:-1]):
                entry.in_queue = index + 1

        # msg = "Starting to execute queue with %d elements: " % len(self.entry_list)
        # for entry in self.entry_list:
        #    msg += str(entry) + " (in_queue=%s) " % entry.in_queue
        # logging.getLogger('queue_exec').info(msg)

    def is_executing(self, node_id=None):
        """
        :returns: True if the queue is executing otherwise False
        :rtype: bool
        """
        status = self._running

        if node_id:
            if self._current_queue_entries:
                for qe in self._current_queue_entries:
                    if qe.get_data_model()._node_id == node_id:
                        status = True
                        break
                    else:
                        status = False
            else:
                status = False

        return status

    def __execute_task(self):
        self._running = True
        # self.emit('centringAllowed', (False, ))
        try:
            for qe in self._queue_entry_list:
                try:
                    self.__execute_entry(qe)
                except (base_queue_entry.QueueAbortedException, Exception) as ex:
                    try:
                        qe.handle_exception(ex)
                        self.stop()
                    except gevent.GreenletExit:
                        pass

                    if isinstance(ex, base_queue_entry.QueueAbortedException):
                        logging.getLogger("user_level_log").warning(
                            "Queue execution was aborted, " + str(ex)
                        )
                    else:
                        logging.getLogger("user_level_log").error(
                            "Queue execution failed with: " + str(ex)
                        )

                    raise ex
        finally:
            self._running = False
            self.emit("queue_execution_finished", (None,))

    def __execute_entry(self, entry):
        if not entry.is_enabled() or self._is_stopped:
            return

        self.emit("queue_entry_execute_started", (entry,))
        self.set_current_entry(entry)
        self._current_queue_entries.append(entry)

        logging.getLogger("queue_exec").info("Executing: " + str(entry))

        if self.is_paused():
            logging.getLogger("user_level_log").info("Queue paused, waiting ...")
            entry.get_view().setText(1, "Queue paused, waiting")

        self.wait_for_pause_event()

        try:
            # Procedure to be done before main implementation
            # of task.
            entry.status = QUEUE_ENTRY_STATUS.RUNNING
            entry.pre_execute()
            entry.execute()

            for child in entry._queue_entry_list:
                self.__execute_entry(child)
            # This part should not be here
            # But somehow exception from collect_failed is not caught here
            if entry.is_failed():
                entry.status = QUEUE_ENTRY_STATUS.FAILED
                self.emit("queue_entry_execute_finished", (entry, "Failed"))
                self.emit(
                    "statusMessage", ("status", "Queue execution failed", "error")
                )
            else:
                entry.status = QUEUE_ENTRY_STATUS.SUCCESS
                self.emit("queue_entry_execute_finished", (entry, "Successful"))
                self.emit("statusMessage", ("status", "", "ready"))
        except base_queue_entry.QueueSkipEntryException as ex:
            logging.getLogger("HWR").warning(
                "encountered Exception (continuing):\n%s" % ex.stack_trace or ex.message
            )
            # Queue entry, failed, skip.
            entry.status = QUEUE_ENTRY_STATUS.SKIPPED
            self.emit("queue_entry_execute_finished", (entry, "Skipped"))
        except base_queue_entry.QueueAbortedException as ex:
            # Queue entry was aborted in a controlled, way.
            # or in the exception case:
            # Definitely not good state, but call post_execute
            # anyway, there might be code that cleans up things
            # done in _pre_execute or before the exception in _execute.
            logging.getLogger("HWR").warning(
                "encountered Exception (continuing):\n%s" % ex.stack_trace or ex.message
            )
            entry.status = QUEUE_ENTRY_STATUS.FAILED
            self.emit("queue_entry_execute_finished", (entry, "Aborted"))
            entry.post_execute()
            entry.handle_exception(ex)
            raise ex
        except base_queue_entry.QueueExecutionException as ex:
            logging.getLogger("HWR").warning(
                "encountered Exception (continuing):\n%s" % ex.stack_trace or ex.message
            )
            entry.status = QUEUE_ENTRY_STATUS.FAILED
            self.emit("queue_entry_execute_finished", (entry, "Failed"))
            self.emit("statusMessage", ("status", "Queue execution failed", "error"))
        except:
            logging.getLogger("HWR").warning(
                "encountered Exception:\n%s" % traceback.format_exc()
            )
            raise
        else:
            entry.post_execute()
        finally:
            # self.emit('queue_entry_execute_finished', (entry, ))
            self.set_current_entry(None)
            self._current_queue_entries.pop(self._current_queue_entries.index(entry))

    def stop(self):
        """
        Stops the queue execution.

        :returns: None
        :rtype: NoneType
        """
        if self._queue_entry_list:
            for qe in self._current_queue_entries:
                try:
                    qe.status = QUEUE_ENTRY_STATUS.FAILED
                    self.emit("queue_entry_execute_finished", (qe, "Aborted"))
                    qe.stop()
                    qe.post_execute()
                except base_queue_entry.QueueAbortedException:
                    pass
                except Exception:
                    pass

        if self._root_task:
            self._root_task.kill(block=False)

        self._queue_end()

    def _queue_end(self):
        # Reset the pause event, in case we were waiting.
        self.set_pause(False)
        self._is_stopped = True
        self._running = False
        self.emit("statusMessage", ("status", "", "Queue stopped"))
        self.emit("queue_stopped", (None,))

    def set_pause(self, state):
        """
        Sets the queue in paused state <state>. Emits the signal queue_paused
        with the current state as parameter.

        :param state: Paused if True running if False
        :type state: bool

        :returns: None
        :rtype: NoneType
        """
        self.emit("queue_paused", (state,))
        self.emit("statusMessage", ("status", "Queue paused", "action_req"))
        # self.emit('centringAllowed', (True, ))
        if state:
            self._paused_event.clear()
        else:
            self._paused_event.set()

    def is_paused(self):
        """
        Returns the pause state, see the method set_pause().

        :returns: None
        :rtype: NoneType
        """
        return not self._paused_event.is_set()

    def pause(self, state):
        """
        Sets the queue in paused state <state> (and waits), paused if True
        running if False.

        :param state: Paused if True running if False
        :type state: bool

        :returns: None
        :rtype: NoneType
        """
        self.set_pause(state)
        self._paused_event.wait()

    def wait_for_pause_event(self):
        """
        Wait for the queue to be set to running set_pause(False) or continue if
        it already was running.

        :returns: None
        :rtype: NoneType
        """
        self._paused_event.wait()

    def disable(self, state):
        """
        Sets the disable state to <state>, disables the possibility
        to call execute if True enables if False.

        :param state: The disabled state, True, False.
        :type state: bool

        :returns: None
        :rtype: NoneType

        """
        self._disable_collect = state

    def is_disabled(self):
        """
        :returns: True if the queue is disabled, (calling execute
                  will do nothing).
        :rtype: bool
        """
        return self._disable_collect

    def set_current_entry(self, entry):
        """
        Sets the currently executing QueueEntry to <entry>.

        :param entry: The entry.
        :type entry: QeueuEntry

        :returns: None
        :rtype: NoneType
        """
        self._current_queue_entry = entry

    def get_current_entry(self):
        """
        Gets the currently executing QueueEntry.

        :returns: The currently executing QueueEntry:
        :rtype: QueueEntry
        """
        return self._current_queue_entry

    def get_entry_with_model(self, model, root_queue_entry=None):
        """
        Find the entry with the data model model.

        :param model: The model to look for.
        :type model: TaskNode

        :returns: The QueueEntry with the model <model>
        :rtype: QueueEntry
        """
        if not root_queue_entry:
            root_queue_entry = self

        for queue_entry in root_queue_entry._queue_entry_list:
            if queue_entry.get_data_model() is model:
                return queue_entry
            else:
                result = self.get_entry_with_model(model, queue_entry)

                if result:
                    return result

    def execute_entry(self, entry, use_async=False):
        """
        Executes the queue entry once the queue has been started <entry>.

        :param entry: The entry to execute.
        :type entry: QueueEntry

        :raises: RuntimeError if the queue is not already running when called
        :returns: None
        :rtype: NoneType
        """
        if not self._running:
            raise RuntimeError(
                "Queue has to be running to execute an entry with execute_entry"
            )

        if use_async:
            gevent.spawn(self.__execute_entry, entry)
        else:
            self.__execute_entry(entry)

    def clear(self):
        """
        Clears the queue (removes all entries).

        :returns: None
        :rtype: NoneType
        """
        self._queue_entry_list = []

    def show_workflow_tab(self):
        self.emit("show_workflow_tab")

    def __str__(self):
        s = "["

        for entry in self._queue_entry_list:
            s += str(entry)

        return s + "]"
